import math
import os
from logging import getLogger
from typing import Any, Dict, List, Optional, Tuple

import rlp
from eth.constants import SECPK1_N
from eth.vm.forks.frontier.transactions import FrontierTransaction
from eth_keys.exceptions import BadSignature
from eth_utils import to_canonical_address, to_checksum_address
from hexbytes import HexBytes
from web3 import Web3
from web3.contract import ContractConstructor

from gnosis.eth.constants import GAS_CALL_DATA_BYTE, NULL_ADDRESS
from gnosis.eth.contracts import (
    get_erc20_contract,
    get_paying_proxy_contract,
    get_safe_V0_0_1_contract,
)
from gnosis.eth.utils import mk_contract_address

logger = getLogger(__name__)


class InvalidERC20Token(Exception):
    pass


class SafeCreationTx:
    def __init__(
        self,
        w3: Web3,
        owners: List[str],
        threshold: int,
        signature_s: int,
        master_copy: str,
        gas_price: int,
        funder: Optional[str],
        payment_token: Optional[str] = None,
        payment_token_eth_value: float = 1.0,
        fixed_creation_cost: Optional[int] = None,
    ):
        """
        Prepare Safe creation
        :param w3: Web3 instance
        :param owners: Owners of the Safe
        :param threshold: Minimum number of users required to operate the Safe
        :param signature_s: Random s value for ecdsa signature
        :param master_copy: Safe master copy address
        :param gas_price: Gas Price
        :param funder: Address to refund when the Safe is created. Address(0) if no need to refund
        :param payment_token: Payment token instead of paying the funder with ether. If None Ether will be used
        :param payment_token_eth_value: Value of payment token per 1 Ether
        :param fixed_creation_cost: Fixed creation cost of Safe (Wei)
        """

        assert 0 < threshold <= len(owners)
        funder = funder or NULL_ADDRESS
        payment_token = payment_token or NULL_ADDRESS
        assert Web3.isChecksumAddress(master_copy)
        assert Web3.isChecksumAddress(funder)
        assert Web3.isChecksumAddress(payment_token)

        self.w3 = w3
        self.owners = owners
        self.threshold = threshold
        self.s = signature_s
        self.master_copy = master_copy
        self.gas_price = gas_price
        self.funder = funder
        self.payment_token = payment_token
        self.payment_token_eth_value = payment_token_eth_value
        self.fixed_creation_cost = fixed_creation_cost

        # Get bytes for `setup(address[] calldata _owners, uint256 _threshold, address to, bytes calldata data)`
        # This initializer will be passed to the proxy and will be called right after proxy is deployed
        safe_setup_data: bytes = self._get_initial_setup_safe_data(owners, threshold)

        # Calculate gas based on experience of previous deployments of the safe
        calculated_gas: int = self._calculate_gas(
            owners, safe_setup_data, payment_token
        )
        # Estimate gas using web3
        estimated_gas: int = self._estimate_gas(
            master_copy, safe_setup_data, funder, payment_token
        )
        self.gas = max(calculated_gas, estimated_gas)

        # Payment will be safe deploy cost + transfer fees for sending ether to the deployer
        self.payment = self._calculate_refund_payment(
            self.gas, gas_price, fixed_creation_cost, payment_token_eth_value
        )

        self.tx_dict: Dict[str, Any] = self._build_proxy_contract_creation_tx(
            master_copy=master_copy,
            initializer=safe_setup_data,
            funder=funder,
            payment_token=payment_token,
            payment=self.payment,
            gas=self.gas,
            gas_price=gas_price,
        )

        self.tx_pyethereum: FrontierTransaction = (
            self._build_contract_creation_tx_with_valid_signature(self.tx_dict, self.s)
        )
        self.tx_raw = rlp.encode(self.tx_pyethereum)
        self.tx_hash = self.tx_pyethereum.hash
        self.deployer_address = to_checksum_address(self.tx_pyethereum.sender)
        self.safe_address = mk_contract_address(self.tx_pyethereum.sender, 0)

        self.v = self.tx_pyethereum.v
        self.r = self.tx_pyethereum.r
        self.safe_setup_data = safe_setup_data

        assert mk_contract_address(self.deployer_address, nonce=0) == self.safe_address

    @property
    def payment_ether(self):
        return self.gas * self.gas_price

    @staticmethod
    def find_valid_random_signature(s: int) -> Tuple[int, int]:
        """
        Find v and r valid values for a given s
        :param s: random value
        :return: v, r
        """
        for _ in range(10000):
            r = int(os.urandom(31).hex(), 16)
            v = (r % 2) + 27
            if r < SECPK1_N:
                tx = FrontierTransaction(0, 1, 21000, b"", 0, b"", v=v, r=r, s=s)
                try:
                    tx.sender
                    return v, r
                except (BadSignature, ValueError):
                    logger.debug("Cannot find signature with v=%d r=%d s=%d", v, r, s)

        raise ValueError("Valid signature not found with s=%d", s)

    @staticmethod
    def _calculate_gas(
        owners: List[str], safe_setup_data: bytes, payment_token: str
    ) -> int:
        """
        Calculate gas manually, based on tests of previosly deployed safes
        :param owners: Safe owners
        :param safe_setup_data: Data for proxy setup
        :param payment_token: If payment token, we will need more gas to transfer and maybe storage if first time
        :return: total gas needed for deployment
        """
        # TODO Do gas calculation estimating the call instead this magic

        base_gas = 60580  # Transaction standard gas

        # If we already have the token, we don't have to pay for storage, so it will be just 5K instead of 20K.
        # The other 1K is for overhead of making the call
        if payment_token != NULL_ADDRESS:
            payment_token_gas = 55000
        else:
            payment_token_gas = 0

        data_gas = GAS_CALL_DATA_BYTE * len(safe_setup_data)  # Data gas
        gas_per_owner = 18020  # Magic number calculated by testing and averaging owners
        return (
            base_gas
            + data_gas
            + payment_token_gas
            + 270000
            + len(owners) * gas_per_owner
        )

    @staticmethod
    def _calculate_refund_payment(
        gas: int,
        gas_price: int,
        fixed_creation_cost: Optional[int],
        payment_token_eth_value: float,
    ) -> int:
        if fixed_creation_cost is None:
            # Payment will be safe deploy cost + transfer fees for sending ether to the deployer
            base_payment: int = (gas + 23000) * gas_price
            # Calculate payment for tokens using the conversion (if used)
            return math.ceil(base_payment / payment_token_eth_value)
        else:
            return fixed_creation_cost

    def _build_proxy_contract_creation_constructor(
        self,
        master_copy: str,
        initializer: bytes,
        funder: str,
        payment_token: str,
        payment: int,
    ) -> ContractConstructor:
        """
        :param master_copy: Master Copy of Gnosis Safe already deployed
        :param initializer: Data initializer to send to GnosisSafe setup method
        :param funder: Address that should get the payment (if payment set)
        :param payment_token: Address if a token is used. If not set, 0x0 will be ether
        :param payment: Payment
        :return: Transaction dictionary
        """
        if not funder or funder == NULL_ADDRESS:
            funder = NULL_ADDRESS
            payment = 0

        return get_paying_proxy_contract(self.w3).constructor(
            master_copy, initializer, funder, payment_token, payment
        )

    def _build_proxy_contract_creation_tx(
        self,
        master_copy: str,
        initializer: bytes,
        funder: str,
        payment_token: str,
        payment: int,
        gas: int,
        gas_price: int,
        nonce: int = 0,
    ):
        """
        :param master_copy: Master Copy of Gnosis Safe already deployed
        :param initializer: Data initializer to send to GnosisSafe setup method
        :param funder: Address that should get the payment (if payment set)
        :param payment_token: Address if a token is used. If not set, 0x0 will be ether
        :param payment: Payment
        :return: Transaction dictionary
        """
        return self._build_proxy_contract_creation_constructor(
            master_copy, initializer, funder, payment_token, payment
        ).buildTransaction(
            {
                "gas": gas,
                "gasPrice": gas_price,
                "nonce": nonce,
            }
        )

    def _build_contract_creation_tx_with_valid_signature(
        self, tx_dict: Dict[str, Any], s: int
    ) -> FrontierTransaction:
        """
        Use pyethereum `Transaction` to generate valid tx using a random signature
        :param tx_dict: Web3 tx dictionary
        :param s: Signature s value
        :return: PyEthereum creation tx for the proxy contract
        """
        zero_address = HexBytes("0x" + "0" * 40)
        f_address = HexBytes("0x" + "f" * 40)
        nonce = tx_dict["nonce"]
        gas_price = tx_dict["gasPrice"]
        gas = tx_dict["gas"]
        to = tx_dict.get("to", b"")  # Contract creation should always have `to` empty
        value = tx_dict["value"]
        data = tx_dict["data"]
        for _ in range(100):
            try:
                v, r = self.find_valid_random_signature(s)
                contract_creation_tx = FrontierTransaction(
                    nonce, gas_price, gas, to, value, HexBytes(data), v=v, r=r, s=s
                )
                sender_address = contract_creation_tx.sender
                contract_address: bytes = to_canonical_address(
                    mk_contract_address(sender_address, nonce)
                )
                if sender_address in (zero_address, f_address) or contract_address in (
                    zero_address,
                    f_address,
                ):
                    raise ValueError("Invalid transaction")
                return contract_creation_tx
            except BadSignature:
                pass
        raise ValueError("Valid signature not found with s=%d", s)

    def _estimate_gas(
        self, master_copy: str, initializer: bytes, funder: str, payment_token: str
    ) -> int:
        """
        Gas estimation done using web3 and calling the node
        Payment cannot be estimated, as no ether is in the address. So we add some gas later.
        :param master_copy: Master Copy of Gnosis Safe already deployed
        :param initializer: Data initializer to send to GnosisSafe setup method
        :param funder: Address that should get the payment (if payment set)
        :param payment_token: Address if a token is used. If not set, 0x0 will be ether
        :return: Total gas estimation
        """

        # Estimate the contract deployment. We cannot estimate the refunding, as the safe address has not any fund
        gas: int = self._build_proxy_contract_creation_constructor(
            master_copy, initializer, funder, payment_token, 0
        ).estimateGas()

        # We estimate the refund as a new tx
        if payment_token == NULL_ADDRESS:
            # Same cost to send 1 ether than 1000
            gas += self.w3.eth.estimate_gas({"to": funder, "value": 1})
        else:
            # Top should be around 52000 when storage is needed (funder no previous owner of token),
            # we use value 1 as we are simulating an internal call, and in that calls you don't pay for the data.
            # If it was a new tx sending 5000 tokens would be more expensive than sending 1 because of data costs
            try:
                gas += (
                    get_erc20_contract(self.w3, payment_token)
                    .functions.transfer(funder, 1)
                    .estimateGas({"from": payment_token})
                )
            except ValueError as exc:
                if "transfer amount exceeds balance" in str(exc):
                    return 70000
                raise InvalidERC20Token from exc

        return gas

    def _get_initial_setup_safe_data(self, owners: List[str], threshold: int) -> bytes:
        return (
            get_safe_V0_0_1_contract(self.w3, self.master_copy)
            .functions.setup(
                owners,
                threshold,
                NULL_ADDRESS,  # Contract address for optional delegate call
                b"",  # Data payload for optional delegate call
            )
            .buildTransaction(
                {
                    "gas": 1,
                    "gasPrice": 1,
                }
            )["data"]
        )



from django.core.exceptions import ValidationError
from django.db import DataError, transaction
from django.test import TestCase

from eth_account import Account
from faker import Faker
from web3 import Web3

from ...constants import NULL_ADDRESS, SENTINEL_ADDRESS
from .models import EthereumAddress, EthereumAddressV2, Keccak256Hash, Sha3Hash, Uint256

faker = Faker()


class TestModels(TestCase):
    def test_ethereum_address_field(self):
        for EthereumAddressModel in (EthereumAddress, EthereumAddressV2):
            with self.subTest(EthereumAddressModel=EthereumAddressModel):
                address = Account.create().address
                self.assertTrue(Web3.isChecksumAddress(address))
                ethereum_address = EthereumAddressModel.objects.create(value=address)
                ethereum_address.refresh_from_db()
                self.assertTrue(Web3.isChecksumAddress(ethereum_address.value))
                self.assertEqual(address, ethereum_address.value)

                # Test addresses
                for addresss in (
                    None,
                    NULL_ADDRESS,
                    SENTINEL_ADDRESS,
                    Account.create().address,
                ):
                    with self.subTest(special_address=addresss):
                        EthereumAddressModel.objects.create(value=addresss)
                        self.assertEqual(
                            EthereumAddressModel.objects.get(value=addresss).value,
                            addresss,
                        )

                with self.assertRaisesMessage(
                    ValidationError,
                    '"0x23" value must be an EIP55 checksummed address.',
                ):
                    with transaction.atomic():
                        EthereumAddressModel.objects.create(value="0x23")

                ethereum_address = EthereumAddressModel(value=Account.create().address)
                self.assertIsNone(ethereum_address.full_clean())

    def test_uint256_field(self):
        for value in [
            2,
            -2,
            2**256,
            2**260,
            25572735541615049941137326092682691158109824779649981270427004917341670006487,
            None,
        ]:
            uint256 = Uint256.objects.create(value=value)
            uint256.refresh_from_db()
            self.assertEqual(uint256.value, value)

        # Overflow
        with self.assertRaises(Exception):
            value = 2**263
            Uint256.objects.create(value=value)

    def test_sha3_hash_field(self):
        value_hexbytes = Web3.keccak(text=faker.name())
        value_hex_with_0x: str = value_hexbytes.hex()
        value_hex_without_0x: str = value_hex_with_0x[2:]
        value: bytes = bytes(value_hexbytes)

        values = [value, value_hex_without_0x, value_hex_with_0x, value_hexbytes]

        for v in values:
            sha3_hash = Sha3Hash.objects.create(value=v)
            sha3_hash.refresh_from_db()
            self.assertEqual(sha3_hash.value, value_hex_with_0x)

        for v in values:
            self.assertEqual(Sha3Hash.objects.filter(value=v).count(), len(values))

        # Hash null
        sha3_hash = Sha3Hash.objects.create(value=None)
        sha3_hash.refresh_from_db()
        self.assertIsNone(sha3_hash.value)

        # Hash too big
        value_hex_invalid: str = "0x" + value_hex_without_0x + "a"
        with self.assertRaisesMessage(
            DataError, "value too long for type character varying(64)"
        ):
            with transaction.atomic():
                Sha3Hash.objects.create(value=value_hex_invalid)

    def test_keccak256_field(self):
        value_hexbytes = Web3.keccak(text=faker.name())
        value_hex_with_0x: str = value_hexbytes.hex()
        value_hex_without_0x: str = value_hex_with_0x[2:]
        value: bytes = bytes(value_hexbytes)

        values = [value, value_hex_without_0x, value_hex_with_0x, value_hexbytes]

        for v in values:
            with self.subTest(v=v):
                keccak256_hash = Keccak256Hash(value=v)
                self.assertIsNone(keccak256_hash.full_clean())
                keccak256_hash.save()
                keccak256_hash.refresh_from_db()
                self.assertEqual(keccak256_hash.value, value_hex_with_0x)

        for v in values:
            with self.subTest(v=v):
                self.assertEqual(
                    Keccak256Hash.objects.filter(value=v).count(), len(values)
                )

        # Hash null
        keccak256_hash = Keccak256Hash.objects.create(value=None)
        keccak256_hash.refresh_from_db()
        self.assertIsNone(keccak256_hash.value)

        # Hash too big
        value_hex_invalid: str = "0x" + value_hex_without_0x + "a"
        with self.assertRaisesMessage(
            ValidationError, f'"{value_hex_invalid}" hash must have exactly 32 bytes.'
        ):
            with transaction.atomic():
                Keccak256Hash.objects.create(value=value_hex_invalid)

        # Hash too small
        value_hex_invalid: str = "0x" + "a1"
        with self.assertRaisesMessage(
            ValidationError, f'"{value_hex_invalid}" hash must have exactly 32 bytes.'
        ):
            with transaction.atomic():
                Keccak256Hash.objects.create(value=value_hex_invalid)

        # Invalid hash
        value_hex_invalid: str = "UX/IO"
        with self.assertRaisesMessage(
            ValidationError,
            f'"{value_hex_invalid}" hash must be a 32 bytes hexadecimal.',
        ):
            with transaction.atomic():
                Keccak256Hash.objects.create(value=value_hex_invalid)



from enum import Enum


class EthereumNetworkNotSupported(Exception):
    pass


class EthereumNetwork(Enum):
    """
    Use https://chainlist.org/ as a reference
    """

    UNKNOWN = -1
    OLYMPIC = 0
    MAINNET = 1
    ROPSTEN = 3
    RINKEBY = 4
    GOERLI = 5
    ETC_KOTTI = 6
    TCH = 7
    UBQ = 8
    OPTIMISTIC = 10
    META = 11
    META_TESTNET = 12
    DIODE_TESTNET = 13
    FLR_FLARE = 14
    DIODE = 15
    FLR_COSTON = 16
    TCH_THAIFI = 17
    TST_TESTNET = 18
    SGB_SONGBIRD = 19
    BOBA_RINKEBY = 28
    RSK = 30
    RSK_TESTNET = 31
    GOOD_TESTNET = 32
    GOOD = 33
    TBWG = 35
    VAL = 38
    TLOS = 40
    TLOS_TESTNET = 41
    KOVAN = 42
    PANGOLIN_FREE_TESTNET = 43
    CRAB_CRAB_NETWORK = 44
    XDC = 50
    TXDC_TESTNET = 51
    CSC = 52
    CSC_TESTNET = 53
    BINANCE = 56
    SYS = 57
    ONTOLOGY = 58
    EOS = 59
    GO = 60
    ELLA = 64
    OKEXCHAIN_TESTNET = 65
    OKEXCHAIN = 66
    DBM_TESTNET = 67
    SOTER = 68
    MIX = 76
    POA_SOKOL = 77
    PC = 78
    GENECHAIN = 80
    METER = 82
    GTTEST_TESTNET = 85
    GT = 86
    TOMO = 88
    EOS_TESTNET = 95
    BSC_CHAPEL = 97
    POA_CORE = 99
    XDAI = 100
    WEB3GAMES_TESTNET = 102
    TT = 108
    XPR_TESTNET = 110
    ETL = 111
    FUSE_MAINNET = 122
    FUSE_SPARK = 123
    DWU = 124
    FETH_FACTORY127 = 127
    HECO = 128
    MATIC = 137
    DAX = 142
    PHT_SIRIUS = 162
    PHT = 163
    RESIL_TESTNET = 172
    AOX_XDAI = 200
    ENERGY_WEB_CHAIN = 246
    FANTOM = 250
    HECO_TESTNET = 256
    HPB = 269
    BOBA = 288
    KCC_TESTNET = 322
    THETA = 361
    THETA_TESTNET_SAPPHIRE = 363
    THETA_TESTNET_AMBER = 364
    THETA_TESTNET = 365
    CRO = 385
    RUPX = 499
    TAO_CORE = 558
    METIS_TESTNET = 588
    MACA_TESTNET = 595
    KAR = 686
    FETH_FACTORY127_TESTNET = 721
    CHEAPETH_CHEAPNET = 777
    ACA = 787
    HAIC = 803
    WAN = 888
    YETI = 977
    WAN_TESTNET = 999
    KLAY_BAOBAB = 1001
    NEW_TESTNET = 1007
    EVC_EVRICE = 1010
    NEW = 1012
    SAKURA = 1022
    CLOVER_TESTNET = 1023
    CLOVER = 1024
    METIS = 1088
    MATH = 1139
    MATH_TESTNET = 1140
    MOON_MOONBEAM = 1284
    MOON_MOONRIVER = 1285
    MOON_MOONROCK = 1286
    MOON_MOONBASE = 1287
    MOON_MOONSHADOW = 1288
    GANACHE = 1337
    DEVNET_3RDSIGHT = 1515
    CATECHAIN = 1618
    EGEM = 1987
    EDG = 2021
    EDG_BERESHEET = 2022
    KORTHO = 2559
    FANTOM_TESTNET = 4002
    IOTEX_IO = 4689
    IOTEX_IO_TESTNET = 4690
    ESN = 5197
    SYS_TESTNET = 5700
    ONTOLOGY_TESTNET = 5851
    RBD = 5869
    SHYFT = 7341
    MDGL_TESTNET = 8029
    GENECHAIN_ADENINE = 8080
    KLAY_CYPRESS = 8217
    KORTHO_TEST = 8285
    OLO = 8723
    OLO_TESTNET = 8724
    BLOXBERG = 8995
    SMARTBCH = 10000
    SMARTBCHTEST_TESTNET = 10001
    GEN = 10101
    SHYFT_TESTNET = 11437
    MTT = 16000
    MTTTEST_DEVNET = 16001
    GO_TESTNET = 31337
    FSN = 32659
    NRG = 39797
    ARBITRUM = 42161
    CELO = 42220
    ATH_ATHEREUM = 43110
    AVALANCHE = 43114
    CELO_ALFAJORES = 44787
    NRG_TESTNET = 49797
    CELO_BAKLAVA = 62320
    VOLTA = 73799
    AKA = 200625
    ARTIS_SIGMA1 = 246529
    ARTIS_TAU1 = 246785
    SPARTA_TESTNET = 333888
    OLYMPUS = 333999
    ARBITRUM_TESTNET = 421611
    ETHO = 1313114
    XERO = 1313500
    MUSIC = 7762959
    MUMBAI = 80001
    PEP_TESTNET = 13371337
    ILT = 18289463
    QKI = 20181205
    AUX = 28945486
    JOYS = 35855456
    AQUA = 61717561
    TOYS_TESTNET = 99415706
    OLT = 311752642
    IPOS = 1122334455
    AURORA = 1313161554
    AURORA_TESTNET = 1313161555
    AURORA_BETANET = 1313161556
    PIRL = 3125659152
    OLT_TESTNET = 4216137055
    PALM_TESTNET = 11297108099
    PALM = 11297108109
    GATHER_DEVNET = 486217935
    GATHER_TESTNET = 356256156
    GATHER_MAINNET = 192837465
    EVMOS_TESTNET = 9000
    EVMOS_MAINNET = 9001
    default = UNKNOWN

    @classmethod
    def _missing_(cls, value):
        return cls.UNKNOWN



class EthereumClientException(ValueError):
    pass


class TransactionAlreadyImported(EthereumClientException):
    pass


class ReplacementTransactionUnderpriced(EthereumClientException):
    pass


class TransactionQueueLimitReached(EthereumClientException):
    pass


class FromAddressNotFound(EthereumClientException):
    pass


class InvalidNonce(EthereumClientException):
    pass


class NonceTooLow(InvalidNonce):
    pass


class NonceTooHigh(InvalidNonce):
    pass


class InsufficientFunds(EthereumClientException):
    pass


class SenderAccountNotFoundInNode(EthereumClientException):
    pass


class UnknownAccount(EthereumClientException):
    pass


class GasLimitExceeded(EthereumClientException):
    pass


class TransactionGasPriceTooLow(EthereumClientException):
    pass


class ParityTraceDecodeException(EthereumClientException):
    pass


class InvalidERC20Info(EthereumClientException):
    pass


class InvalidERC721Info(EthereumClientException):
    pass


class BatchCallException(EthereumClientException):
    pass


class BatchCallFunctionFailed(BatchCallException):
    pass



from typing import Any, Dict, List, Sequence
from unittest import mock
from unittest.mock import MagicMock

from django.test import TestCase

from eth_account import Account
from hexbytes import HexBytes
from web3.eth import Eth
from web3.types import TxParams

from ..constants import GAS_CALL_DATA_BYTE, NULL_ADDRESS
from ..contracts import get_erc20_contract
from ..ethereum_client import (
    EthereumClient,
    EthereumClientProvider,
    EthereumNetwork,
    FromAddressNotFound,
    InsufficientFunds,
    InvalidNonce,
    ParityManager,
    SenderAccountNotFoundInNode,
)
from ..exceptions import BatchCallException, InvalidERC20Info
from ..utils import get_eth_address_with_key
from .ethereum_test_case import EthereumTestCaseMixin
from .mocks.mock_internal_txs import creation_internal_txs, internal_txs_errored
from .mocks.mock_log_receipts import invalid_log_receipt, log_receipts
from .mocks.mock_trace_block import trace_block_2191709_mock, trace_block_13191781_mock
from .mocks.mock_trace_filter import trace_filter_mock_1
from .mocks.mock_trace_transaction import trace_transaction_mocks
from .utils import just_test_if_mainnet_node


class TestERC20Module(EthereumTestCaseMixin, TestCase):
    def test_decode_transfer_log(self):
        decoded_logs = self.ethereum_client.erc20.decode_logs(
            log_receipts + [invalid_log_receipt]
        )
        self.assertEqual(len(decoded_logs), 6)
        self.assertEqual(
            len([event for event in decoded_logs if "tokenId" in event["args"]]), 2
        )
        self.assertEqual(
            len([event for event in decoded_logs if "value" in event["args"]]), 3
        )
        self.assertEqual(
            len([event for event in decoded_logs if "unknown" in event["args"]]), 1
        )

        expected_log_1 = {
            "address": "0x39C4BFa00b6edecCDd00fA9589E1BE76DE63e862",
            "topics": [
                HexBytes(
                    "0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef"
                ),
                HexBytes(
                    "0x00000000000000000000000094e01661ebaef430fe862f958c03200b0f483f27"
                ),
                HexBytes(
                    "0x00000000000000000000000064da772dd84965f0ee58174941d78a9dfbccca2e"
                ),
            ],
            "data": "0x000000000000000000000000000000000000000000000001a055690d9db80000",
            "blockNumber": 4357126,
            "transactionHash": HexBytes(
                "0x21381484d8f69dcd782560d1fd3cd818e743c79767985d01aec7e61c2a7f1de9"
            ),
            "transactionIndex": 14,
            "blockHash": HexBytes(
                "0x677ada1a306fc50751001bca6eeaa3f5a87a0bf2c9f6fa27899bfbaf999cca4f"
            ),
            "logIndex": 14,
            "removed": False,
            "args": {
                "from": "0x94E01661eBaef430fE862f958c03200b0F483f27",
                "to": "0x64DA772DD84965f0Ee58174941d78a9DfBccca2e",
                "value": 30000000000000000000,
            },
        }
        expected_log_3 = {
            "address": "0x99b9F9BA62002a9b43aF6e540428277D5E52EF47",
            "topics": [
                HexBytes(
                    "0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef"
                ),
                HexBytes(
                    "0x00000000000000000000000099b9f9ba62002a9b43af6e540428277d5e52ef47"
                ),
                HexBytes(
                    "0x00000000000000000000000094e01661ebaef430fe862f958c03200b0f483f27"
                ),
                HexBytes(
                    "0xcc292d3dab2c0fbbf616670ac57ec51162959c2d9cbe938819b6e8bc1c757335"
                ),
            ],
            "data": "0x",
            "blockNumber": 4357126,
            "transactionHash": HexBytes(
                "0x21381484d8f69dcd782560d1fd3cd818e743c79767985d01aec7e61c2a7f1de9"
            ),
            "transactionIndex": 14,
            "blockHash": HexBytes(
                "0x677ada1a306fc50751001bca6eeaa3f5a87a0bf2c9f6fa27899bfbaf999cca4f"
            ),
            "logIndex": 17,
            "removed": False,
            "args": {
                "from": "0x99b9F9BA62002a9b43aF6e540428277D5E52EF47",
                "to": "0x94E01661eBaef430fE862f958c03200b0F483f27",
                "tokenId": 92344574081811136966607054691835999266725413506859602442775390826469350798133,
            },
        }

        expected_log_5 = {
            "address": "0xe35F3B71CA90eE2606F64b645D8F4f8DCaA914Bf",
            "topics": [
                HexBytes(
                    "0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef"
                )
            ],
            "data": "0x0000000000000000000000008683f9c4e856be65f8a38a3a768e8fd6de94d30a000000000000000000000000fd1017c3284a12ac33bc65df12d71721c85931e00000000000000000000000000000000000000000000000000000000000000001",
            "blockNumber": 7240557,
            "transactionHash": HexBytes(
                "0x1962e296457b16d5221d33623f2db5f617cb54221deb7cfd73611f761ac526a3"
            ),
            "transactionIndex": 6,
            "blockHash": HexBytes(
                "0xdf3c33d034f1b342820afdfb2612d0794d4a1c15518184d71a047ef1eb151d10"
            ),
            "logIndex": 5,
            "removed": False,
            "args": {
                "from": "0x8683f9c4e856be65f8a38a3a768e8fd6de94d30a",
                "to": "0xfd1017c3284a12ac33bc65df12d71721c85931e0",
                "unknown": 1,
            },
        }

        self.assertEqual(decoded_logs[1], expected_log_1)
        self.assertEqual(decoded_logs[3], expected_log_3)
        self.assertEqual(decoded_logs[5], expected_log_5)

    def test_decode_invalid_transfer_log(self):
        invalid_transfer_logs = [invalid_log_receipt]
        self.assertEqual(
            self.ethereum_client.erc20.decode_logs(invalid_transfer_logs), []
        )

    def test_get_name_symbol_balance(self):
        amount = 1000
        address = Account.create().address
        erc20_contract = self.deploy_example_erc20(amount, address)
        token_balance = self.ethereum_client.erc20.get_balance(
            address, erc20_contract.address
        )
        self.assertTrue(token_balance, amount)

        another_account = Account.create().address
        token_balance = self.ethereum_client.erc20.get_balance(
            another_account, erc20_contract.address
        )
        self.assertEqual(token_balance, 0)

        self.assertTrue(self.ethereum_client.erc20.get_name(erc20_contract.address))
        self.assertTrue(self.ethereum_client.erc20.get_symbol(erc20_contract.address))

    def test_get_balances(self):
        account_address = Account.create().address
        self.assertEqual(
            self.ethereum_client.erc20.get_balances(account_address, []),
            [{"token_address": None, "balance": 0}],
        )

        value = 7
        self.send_ether(account_address, 7)
        self.assertEqual(
            self.ethereum_client.erc20.get_balances(account_address, []),
            [{"token_address": None, "balance": value}],
        )

        tokens_value = 12
        erc20 = self.deploy_example_erc20(tokens_value, account_address)
        self.assertCountEqual(
            self.ethereum_client.erc20.get_balances(account_address, [erc20.address]),
            [
                {"token_address": None, "balance": value},
                {"token_address": erc20.address, "balance": tokens_value},
            ],
        )

        tokens_value_2 = 19
        erc20_2 = self.deploy_example_erc20(tokens_value_2, account_address)
        self.assertCountEqual(
            self.ethereum_client.erc20.get_balances(
                account_address, [erc20.address, erc20_2.address]
            ),
            [
                {"token_address": None, "balance": value},
                {"token_address": erc20.address, "balance": tokens_value},
                {"token_address": erc20_2.address, "balance": tokens_value_2},
            ],
        )

        with mock.patch.object(
            EthereumClient,
            "batch_call_same_function",
            return_value=[
                b"\x08\xc3y\xa0\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00 \x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x17Only the proxy can call\x00\x00\x00\x00\x00\x00\x00\x00\x00",
                5,
            ],
        ):
            token_addresses = [
                "0x57Ab1E02fEE23774580C119740129eAC7081e9D3",
                "0x6810e776880C02933D47DB1b9fc05908e5386b96",
            ]
            self.assertCountEqual(
                self.ethereum_client.erc20.get_balances(
                    account_address, token_addresses
                ),
                [
                    {"token_address": None, "balance": value},
                    {
                        "token_address": "0x57Ab1E02fEE23774580C119740129eAC7081e9D3",
                        "balance": 0,
                    },
                    {
                        "token_address": "0x6810e776880C02933D47DB1b9fc05908e5386b96",
                        "balance": 5,
                    },
                ],
            )

    def test_get_total_transfer_history(self):
        amount = 50
        owner_account = self.create_account(initial_ether=0.01)
        account_1 = self.create_account(initial_ether=0.01)
        account_2 = self.create_account(initial_ether=0.01)
        account_3 = self.create_account(initial_ether=0.01)
        erc20_contract = self.deploy_example_erc20(amount, owner_account.address)
        # `owner` sends `amount // 2` to `account_1` and `account_3`
        self.send_tx(
            erc20_contract.functions.transfer(
                account_1.address, amount // 2
            ).buildTransaction({"from": owner_account.address}),
            owner_account,
        )
        self.send_tx(
            erc20_contract.functions.transfer(
                account_3.address, amount // 2
            ).buildTransaction({"from": owner_account.address}),
            owner_account,
        )
        logs = self.ethereum_client.erc20.get_total_transfer_history(
            [account_1.address]
        )
        self.assertEqual(len(logs), 1)
        self.assertEqual(logs[0]["args"]["from"], owner_account.address)
        self.assertEqual(logs[0]["args"]["to"], account_1.address)
        self.assertEqual(logs[0]["args"]["value"], amount // 2)
        log_0_block_number = logs[0]["blockNumber"]

        # `account1` sends `amount // 2` (all) to `account_2`
        self.send_tx(
            erc20_contract.functions.transfer(
                account_2.address, amount // 2
            ).buildTransaction({"from": account_1.address}),
            account_1,
        )
        # Test `token_address` and `to_block` parameters
        logs = self.ethereum_client.erc20.get_total_transfer_history(
            [account_1.address], token_address=erc20_contract.address, to_block="latest"
        )
        self.assertEqual(len(logs), 2)
        self.assertEqual(logs[1]["args"]["from"], account_1.address)
        self.assertEqual(logs[1]["args"]["to"], account_2.address)
        self.assertEqual(logs[1]["args"]["value"], amount // 2)
        self.assertGreaterEqual(logs[1]["blockNumber"], log_0_block_number)

        logs = self.ethereum_client.erc20.get_total_transfer_history(
            [account_3.address]
        )
        self.assertEqual(len(logs), 1)
        self.assertEqual(logs[0]["args"]["from"], owner_account.address)
        self.assertEqual(logs[0]["args"]["to"], account_3.address)
        self.assertEqual(logs[0]["args"]["value"], amount // 2)

        logs = self.ethereum_client.erc20.get_total_transfer_history(
            [account_2.address, account_3.address]
        )
        self.assertEqual(len(logs), 2)

        logs = self.ethereum_client.erc20.get_total_transfer_history()
        self.assertGreaterEqual(len(logs), 3)

    def test_get_transfer_history(self):
        amount = 1000
        owner_account = self.create_account(initial_ether=0.01)

        # Owner will send amount / 2 to receiver and receiver2. Then receiver1 and receiver 2
        # will send amount / 4 to receiver3
        receiver_account = self.create_account(initial_ether=0.01)
        receiver2_account = self.create_account(initial_ether=0.01)
        receiver3_account = self.create_account(initial_ether=0.01)
        erc20_contract = self.deploy_example_erc20(amount, owner_account.address)
        block_number = self.w3.eth.block_number
        events = self.ethereum_client.erc20.get_transfer_history(
            block_number, token_address=erc20_contract.address
        )
        self.assertEqual(len(events), 1)  # Event is triggered on minting

        self.send_tx(
            erc20_contract.functions.transfer(
                receiver_account.address, amount // 2
            ).buildTransaction({"from": owner_account.address}),
            owner_account,
        )
        self.send_tx(
            erc20_contract.functions.transfer(
                receiver2_account.address, amount // 2
            ).buildTransaction({"from": owner_account.address}),
            owner_account,
        )

        self.send_tx(
            erc20_contract.functions.transfer(
                receiver3_account.address, amount // 4
            ).buildTransaction({"from": receiver_account.address}),
            receiver_account,
        )
        self.send_tx(
            erc20_contract.functions.transfer(
                receiver3_account.address, amount // 4
            ).buildTransaction({"from": receiver2_account.address}),
            receiver2_account,
        )

        events = self.ethereum_client.erc20.get_transfer_history(
            block_number, token_address=erc20_contract.address
        )
        self.assertEqual(len(events), 5)

        events = self.ethereum_client.erc20.get_transfer_history(
            block_number, from_address=owner_account.address
        )
        self.assertEqual(len(events), 2)

        events = self.ethereum_client.erc20.get_transfer_history(
            block_number, from_address=receiver_account.address
        )
        self.assertEqual(len(events), 1)

        events = self.ethereum_client.erc20.get_transfer_history(
            block_number, from_address=receiver2_account.address
        )
        self.assertEqual(len(events), 1)

        events = self.ethereum_client.erc20.get_transfer_history(
            block_number, from_address=receiver3_account.address
        )
        self.assertEqual(len(events), 0)

        events = self.ethereum_client.erc20.get_transfer_history(
            block_number, to_address=receiver2_account.address
        )
        self.assertEqual(len(events), 1)

        events = self.ethereum_client.erc20.get_transfer_history(
            block_number, to_address=receiver3_account.address
        )
        self.assertEqual(len(events), 2)
        for event in events:
            self.assertEqual(event.args.value, amount // 4)
            self.assertEqual(event.args.to, receiver3_account.address)

        events = self.ethereum_client.erc20.get_transfer_history(
            block_number,
            from_address=receiver2_account.address,
            to_address=receiver3_account.address,
        )
        self.assertEqual(len(events), 1)
        event = events[0]
        self.assertEqual(event.args.value, amount // 4)
        self.assertEqual(event.args["from"], receiver2_account.address)
        self.assertEqual(event.args.to, receiver3_account.address)

    def test_get_info(self):
        amount = 1
        owner = Account.create().address
        erc20_contract = self.deploy_example_erc20(amount, owner)
        erc20_info = self.ethereum_client.erc20.get_info(erc20_contract.address)
        self.assertEqual(erc20_info.decimals, 18)

        with self.assertRaises(InvalidERC20Info):
            self.ethereum_client.erc20.get_info(Account.create().address)

    def test_send_tokens(self):
        amount = 5
        owner = self.ethereum_test_account
        owner_2 = Account.create()
        erc20 = self.deploy_example_erc20(amount, owner.address)
        self.assertEqual(
            self.ethereum_client.erc20.get_balance(owner.address, erc20.address), amount
        )

        amount_2 = 3
        self.ethereum_client.erc20.send_tokens(
            owner_2.address, amount_2, erc20.address, owner.key
        )
        self.assertEqual(
            self.ethereum_client.erc20.get_balance(owner.address, erc20.address),
            amount - amount_2,
        )
        self.assertEqual(
            self.ethereum_client.erc20.get_balance(owner_2.address, erc20.address),
            amount_2,
        )


class TestParityManager(EthereumTestCaseMixin, TestCase):
    def test_decode_trace(self):
        example_traces = [
            {
                "action": {
                    "callType": "call",
                    "from": "0x32be343b94f860124dc4fee278fdcbd38c102d88",
                    "gas": "0x4c40d",
                    "input": "0x",
                    "to": "0x8bbb73bcb5d553b5a556358d27625323fd781d37",
                    "value": "0x3f0650ec47fd240000",
                },
                "blockHash": "0x86df301bcdd8248d982dbf039f09faf792684e1aeee99d5b58b77d620008b80f",
                "blockNumber": 3068183,
                "result": {"gasUsed": "0x0", "output": "0x"},
                "subtraces": 0,
                "traceAddress": [],
                "transactionHash": "0x3321a7708b1083130bd78da0d62ead9f6683033231617c9d268e2c7e3fa6c104",
                "transactionPosition": 3,
                "type": "call",
            },
            {
                "action": {
                    "from": "0x3b169a0fb55ea0b6bafe54c272b1fe4983742bf7",
                    "gas": "0x49b0b",
                    "init": " ",
                    "value": "0x0",
                },
                "blockHash": "0x03f9f64dfeb7807b5df608e6957dd4d521fd71685aac5533451d27f0abe03660",
                "blockNumber": 3793534,
                "result": {
                    "address": "0x61a7cc907c47c133d5ff5b685407201951fcbd08",
                    "code": "0x60806040526004361061004c576000357c0100000000000000000000000000000000000000000000000000000000900463ffffffff1680634555d5c91461008b5780635c60da1b146100b6575b73ffffffffffffffffffffffffffffffffffffffff600054163660008037600080366000845af43d6000803e6000811415610086573d6000fd5b3d6000f35b34801561009757600080fd5b506100a061010d565b6040518082815260200191505060405180910390f35b3480156100c257600080fd5b506100cb610116565b604051808273ffffffffffffffffffffffffffffffffffffffff1673ffffffffffffffffffffffffffffffffffffffff16815260200191505060405180910390f35b60006002905090565b60008060009054906101000a900473ffffffffffffffffffffffffffffffffffffffff169050905600a165627a7a7230582007fffd557dfc8c4d2fdf56ba6381a6ce5b65b6260e1492d87f26c6d4f1d041080029",
                    "gasUsed": "0x4683f",
                },
                "subtraces": 2,
                "traceAddress": [],
                "transactionHash": "0x6c7e8f8778d33d81b29c4bd7526ee50a4cea340d69eed6c89ada4e6fab731789",
                "transactionPosition": 1,
                "type": "create",
            },
            {
                "action": {
                    "callType": "call",
                    "from": "0xed85033a23027439fb20f381f5930ba67f1ebee0",
                    "gas": "0x0",
                    "input": "0x",
                    "to": "0x0affccd762f0a7fa1bca40c51afe1a806a74a6f9",
                    "value": "0x11c290633b7000",
                },
                "blockHash": "0xbf3a508137e72b20399852becdee1752c7b9986184f835014d12b10c01746f39",
                "blockNumber": 7556887,
                "error": "Out of gas",
                "subtraces": 0,
                "traceAddress": [],
                "transactionHash": "0xec7e447ce8eef033a8c85442a281bd34436e576553c7d98fbe859af7754a9064",
                "transactionPosition": 117,
                "type": "call",
            },
            {
                "action": {
                    "address": "0x4440adafbc6c4e45c299451c0eedc7c8b98c14ac",
                    "balance": "0xa",
                    "refundAddress": "0x1240adafbc6c4e45c299451c0eedc7c8b98c2222",
                },
                "blockHash": "0x8512d367492371edf44ebcbbbd935bc434946dddc2b126cb558df5906012186c",
                "blockNumber": 7829689,
                "result": None,
                "subtraces": 0,
                "traceAddress": [0, 0, 0, 0, 0, 0],
                "transactionHash": "0x5f7af6aa390f9f8dd79ee692c37cbde76bb7869768b1bac438b6d176c94f637d",
                "transactionPosition": 35,
                "type": "suicide",
            },
        ]
        at_least_one_error = False
        at_least_one_self_destruct = False
        decoded_traces = self.ethereum_client.parity._decode_traces(example_traces)
        for example_trace, decoded_trace in zip(example_traces, decoded_traces):
            if decoded_trace["type"] == "suicide":
                self.assertEqual(
                    decoded_trace["action"]["address"],
                    "0x4440AdaFBc6c4E45C299451C0eEdC7C8B98c14Ac",
                )
                self.assertEqual(decoded_trace["action"]["balance"], 10)
                self.assertEqual(
                    decoded_trace["action"]["refundAddress"],
                    "0x1240aDafBC6C4e45C299451C0eEdC7c8B98C2222",
                )
                at_least_one_self_destruct = True
            else:
                self.assertEqual(
                    decoded_trace["action"]["gas"],
                    int(example_trace["action"]["gas"], 16),
                )
                self.assertEqual(
                    decoded_trace["action"]["value"],
                    int(example_trace["action"]["value"], 16),
                )
                if "error" in decoded_trace:
                    self.assertNotIn("result", decoded_trace)
                    at_least_one_error = True
                else:
                    self.assertEqual(
                        decoded_trace["result"]["gasUsed"],
                        int(example_trace["result"]["gasUsed"], 16),
                    )

        self.assertTrue(at_least_one_error)
        self.assertTrue(at_least_one_self_destruct)

        self.assertEqual(decoded_traces[0]["result"]["output"], HexBytes(""))
        self.assertEqual(
            decoded_traces[1]["result"]["address"],
            self.w3.toChecksumAddress(example_traces[1]["result"]["address"]),
        )
        self.assertEqual(
            decoded_traces[1]["result"]["code"],
            HexBytes(example_traces[1]["result"]["code"]),
        )

        self.assertEqual(decoded_traces[2]["error"], "Out of gas")

    def test_filter_out_errored_traces(self):
        self.assertEqual(self.ethereum_client.parity.filter_out_errored_traces([]), [])
        traces = internal_txs_errored
        expected = [
            {
                "action": {
                    "from": "0x667dEb5A98f77052cf561658575cF1530Ee42C7a",
                    "gas": 60066,
                    "value": 0,
                    "callType": "call",
                    "input": HexBytes(
                        " "
                    ),
                    "to": "0x47F61944efdB020829caead65AfF8AC024600580",
                },
                "blockHash": "0x4c49052fc99be82b91f8a35320826304dfe278dbd7d756edde000d331606358f",
                "blockNumber": 4735890,
                "result": {
                    "gasUsed": 21838,
                    "output": HexBytes(
                        "0x0000000000000000000000000000000000000000000000000000000000000000"
                    ),
                },
                "subtraces": 1,
                "traceAddress": [],
                "transactionHash": "0xf097d5e5dd39a6799fc13dfa49732a115b457386520dc92f99f0135a1d196851",
                "transactionPosition": 3,
                "type": "call",
            },
            {
                "action": {
                    "from": "0x47F61944efdB020829caead65AfF8AC024600580",
                    "gas": 57726,
                    "value": 0,
                    "callType": "delegatecall",
                    "input": HexBytes(
                        " "
                    ),
                    "to": "0xb6029EA3B2c51D09a50B53CA8012FeEB05bDa35A",
                },
                "blockHash": "0x4c49052fc99be82b91f8a35320826304dfe278dbd7d756edde000d331606358f",
                "blockNumber": 4735890,
                "result": {
                    "gasUsed": 20369,
                    "output": HexBytes(
                        "0x0000000000000000000000000000000000000000000000000000000000000000"
                    ),
                },
                "subtraces": 1,
                "traceAddress": [0],
                "transactionHash": "0xf097d5e5dd39a6799fc13dfa49732a115b457386520dc92f99f0135a1d196851",
                "transactionPosition": 3,
                "type": "call",
            },
        ]
        self.assertEqual(
            self.ethereum_client.parity.filter_out_errored_traces(traces), expected
        )

        traces = [
            {
                "traceAddress": [],
            },
            {
                "traceAddress": [0],
            },
            {
                "traceAddress": [0, 0],
            },
            {
                "error": "reverted",
                "traceAddress": [0, 0, 0],
            },
            {
                "traceAddress": [0, 0, 0, 0],
            },
            {
                "traceAddress": [0, 0, 0, 1],
            },
            {
                "traceAddress": [0, 0, 1],
            },
            {
                "traceAddress": [0, 0, 1, 0],
            },
            {
                "error": "reverted",
                "traceAddress": [0, 0, 1, 0, 0],
            },
            {
                "traceAddress": [0, 0, 1, 0, 1],
            },
            {
                "traceAddress": [0, 0, 1, 0, 2],
            },
            {
                "error": "reverted",
                "traceAddress": [0, 0, 1, 0, 3],
            },
            {
                "traceAddress": [0, 0, 2],
            },
        ]
        expected = [
            {
                "traceAddress": [],
            },
            {
                "traceAddress": [0],
            },
            {
                "traceAddress": [0, 0],
            },
            {
                "traceAddress": [0, 0, 1],
            },
            {
                "traceAddress": [0, 0, 1, 0],
            },
            {
                "traceAddress": [0, 0, 1, 0, 1],
            },
            {
                "traceAddress": [0, 0, 1, 0, 2],
            },
            {
                "traceAddress": [0, 0, 2],
            },
        ]
        self.assertEqual(
            sorted(traces, key=lambda trace: trace["traceAddress"]), traces
        )
        self.assertEqual(
            self.ethereum_client.parity.filter_out_errored_traces(traces), expected
        )

    @mock.patch.object(
        ParityManager, "trace_transaction", return_value=internal_txs_errored
    )
    def test_get_previous_trace(self, trace_transaction_mock: MagicMock):
        self.assertEqual(
            self.ethereum_client.parity.get_previous_trace("0x12", [0, 0])[
                "traceAddress"
            ],
            [0],
        )
        self.assertEqual(
            self.ethereum_client.parity.get_previous_trace(
                "0x12", [0, 0], number_traces=2
            )["traceAddress"],
            [],
        )
        self.assertEqual(
            self.ethereum_client.parity.get_previous_trace(
                "0x12", [0, 0], skip_delegate_calls=True
            )["traceAddress"],
            [],
        )
        self.assertIsNone(
            self.ethereum_client.parity.get_previous_trace(
                "0x12", [0, 0], number_traces=3
            )
        )
        self.assertEqual(
            self.ethereum_client.parity.get_previous_trace(
                "0x12", [0, 0, 0], skip_delegate_calls=True
            )["traceAddress"],
            [0, 0],
        )

    @mock.patch.object(
        ParityManager, "trace_transaction", return_value=creation_internal_txs
    )
    def test_get_next_traces(self, trace_transaction_mock: MagicMock):
        def trace_addresses(traces: Sequence[Dict[str, Any]]) -> List[List[int]]:
            return [trace["traceAddress"] for trace in traces]

        self.assertEqual(
            trace_addresses(self.ethereum_client.parity.get_next_traces("0x12", [])),
            [[0], [1]],
        )
        self.assertEqual(
            trace_addresses(
                self.ethereum_client.parity.get_next_traces(
                    "0x12", [], remove_delegate_calls=True
                )
            ),
            [[1]],
        )
        self.assertEqual(
            trace_addresses(
                self.ethereum_client.parity.get_next_traces(
                    "0x12", [], remove_calls=True
                )
            ),
            [[0]],
        )
        self.assertEqual(
            trace_addresses(
                self.ethereum_client.parity.get_next_traces(
                    "0x12", [], remove_delegate_calls=True, remove_calls=True
                )
            ),
            [],
        )
        self.assertEqual(self.ethereum_client.parity.get_next_traces("0x12", [0]), [])
        self.assertEqual(
            trace_addresses(self.ethereum_client.parity.get_next_traces("0x12", [1])),
            [[1, 0]],
        )

    def test_trace_filter(self):
        with self.assertRaisesMessage(AssertionError, "at least"):
            self.ethereum_client.parity.trace_filter()

        with self.assertRaisesMessage(ValueError, "Method trace_filter not supported"):
            self.ethereum_client.parity.trace_filter(
                from_address=Account.create().address
            )


class TestEthereumNetwork(EthereumTestCaseMixin, TestCase):
    def test_default_ethereum_network_name(self):
        self.assertEqual(
            EthereumNetwork(EthereumNetwork.default), EthereumNetwork.UNKNOWN
        )

    def test_unknown_ethereum_network_name(self):
        self.assertEqual(EthereumNetwork(2), EthereumNetwork.UNKNOWN)

    def test_mainnet_ethereum_network_name(self):
        self.assertEqual(EthereumNetwork(1), EthereumNetwork.MAINNET)

    def test_rinkeby_ethereum_network_name(self):
        self.assertEqual(EthereumNetwork(4), EthereumNetwork.RINKEBY)


class TestEthereumClient(EthereumTestCaseMixin, TestCase):
    def test_ethereum_client_str(self):
        self.assertTrue(str(self.ethereum_client))

    def test_current_block_number(self):
        self.assertGreaterEqual(self.ethereum_client.current_block_number, 0)

    def test_get_transaction(self):
        to = Account.create().address
        value = 123
        tx_hash = self.send_ether(to, value)
        tx = self.ethereum_client.get_transaction(tx_hash)
        self.assertEqual(tx.to, to)
        self.assertEqual(tx.value, value)
        block = self.ethereum_client.get_block(tx.blockNumber)
        self.assertEqual(block.number, tx.blockNumber)

    def test_get_transactions(self):
        self.assertEqual(self.ethereum_client.get_transactions([]), [])
        to = Account.create().address
        values = [123, 234, 567]
        tx_hashes = [self.send_ether(to, values[i]) for i in range(3)]
        txs = self.ethereum_client.get_transactions(tx_hashes)
        for i, tx in enumerate(txs):
            self.assertEqual(tx["to"], to)
            self.assertEqual(tx["value"], values[i])

        self.assertEqual(self.ethereum_client.get_transaction_receipts([]), [])
        receipts = self.ethereum_client.get_transaction_receipts(tx_hashes)
        for i, receipt in enumerate(receipts):
            self.assertEqual(receipt["status"], 1)
            self.assertGreaterEqual(receipt["gasUsed"], 21000)

    def test_check_tx_with_confirmations(self):
        value = 1
        to, _ = get_eth_address_with_key()

        tx_hash = self.ethereum_client.send_eth_to(
            self.ethereum_test_account.key, to=to, gas_price=self.gas_price, value=value
        )
        self.assertFalse(self.ethereum_client.check_tx_with_confirmations(tx_hash, 2))

        _ = self.ethereum_client.send_eth_to(
            self.ethereum_test_account.key, to=to, gas_price=self.gas_price, value=value
        )
        self.assertFalse(self.ethereum_client.check_tx_with_confirmations(tx_hash, 2))

        _ = self.ethereum_client.send_eth_to(
            self.ethereum_test_account.key, to=to, gas_price=self.gas_price, value=value
        )
        self.assertTrue(self.ethereum_client.check_tx_with_confirmations(tx_hash, 2))

    def test_estimate_gas(self):
        send_ether_gas = 21000
        from_ = self.ethereum_test_account.address
        to, _ = get_eth_address_with_key()
        gas = self.ethereum_client.estimate_gas(
            to, from_=from_, value=5, data=None, block_identifier="pending"
        )
        self.assertEqual(gas, send_ether_gas)

        gas = self.ethereum_client.estimate_gas(to, from_=from_, value=5, data=b"")
        self.assertEqual(gas, send_ether_gas)

        gas = self.ethereum_client.estimate_gas(to, from_=from_, value=5, data=None)
        self.assertEqual(gas, send_ether_gas)

    def test_estimate_gas_with_erc20(self):
        send_ether_gas = 21000
        amount_tokens = 5000
        amount_to_send = amount_tokens // 2
        from_account = self.ethereum_test_account
        from_ = from_account.address
        to, _ = get_eth_address_with_key()

        erc20_contract = self.deploy_example_erc20(amount_tokens, from_)
        transfer_tx = erc20_contract.functions.transfer(
            to, amount_to_send
        ).buildTransaction({"from": from_})
        data = transfer_tx["data"]

        # Ganache does not cares about all this anymore
        # ---------------------------------------------
        # Ganache returns error because there is data but address is not a contract
        # with self.assertRaisesMessage(ValueError, 'transaction which calls a contract function'):
        #    self.ethereum_client.estimate_gas(from_, to, 0, data, block_identifier='pending')
        # `to` is not a contract, no functions will be triggered
        # gas = self.ethereum_client.estimate_gas(from_, to, 0, data, block_identifier='pending')
        # self.assertGreater(gas, send_ether_gas)
        # self.assertLess(gas, send_ether_gas + 2000)

        gas = self.ethereum_client.estimate_gas(
            erc20_contract.address,
            from_=from_,
            value=0,
            data=data,
            block_identifier="pending",
        )
        self.assertGreater(gas, send_ether_gas)

        # We do the real erc20 transfer for the next test case
        self.send_tx(transfer_tx, from_account)
        token_balance = self.ethereum_client.erc20.get_balance(
            to, erc20_contract.address
        )
        self.assertTrue(token_balance, amount_to_send)

        # Gas will be lowest now because you don't have to pay for storage
        gas2 = self.ethereum_client.estimate_gas(
            erc20_contract.address,
            from_=from_,
            value=0,
            data=data,
            block_identifier="pending",
        )
        self.assertLess(gas2, gas)

    def test_get_ethereum_network(self):
        self.assertEqual(self.ethereum_client.get_network(), EthereumNetwork.GANACHE)

        with mock.patch.object(
            Eth, "chain_id", return_value=1, new_callable=mock.PropertyMock
        ):
            # Test caching
            self.assertEqual(
                self.ethereum_client.get_network(), EthereumNetwork.GANACHE
            )
            self.ethereum_client.get_network.cache_clear()
            self.assertEqual(
                self.ethereum_client.get_network(), EthereumNetwork.MAINNET
            )
            self.ethereum_client.get_network.cache_clear()

        with mock.patch.object(
            Eth, "chain_id", return_value=4, new_callable=mock.PropertyMock
        ):
            self.assertEqual(
                self.ethereum_client.get_network(), EthereumNetwork.RINKEBY
            )
            self.ethereum_client.get_network.cache_clear()

        with mock.patch.object(
            Eth, "chain_id", return_value=4815162342, new_callable=mock.PropertyMock
        ):
            self.assertEqual(
                self.ethereum_client.get_network(), EthereumNetwork.UNKNOWN
            )
            self.ethereum_client.get_network.cache_clear()

    def test_get_nonce(self):
        address = Account.create().address
        nonce = self.ethereum_client.get_nonce_for_account(address)
        self.assertEqual(nonce, 0)

        nonce = self.ethereum_client.get_nonce_for_account(
            address, block_identifier="pending"
        )
        self.assertEqual(nonce, 0)

    def test_estimate_data_gas(self):
        self.assertEqual(self.ethereum_client.estimate_data_gas(HexBytes("")), 0)
        self.assertEqual(self.ethereum_client.estimate_data_gas(HexBytes("0x00")), 4)
        self.assertEqual(
            self.ethereum_client.estimate_data_gas(HexBytes("0x000204")),
            4 + GAS_CALL_DATA_BYTE * 2,
        )
        self.assertEqual(
            self.ethereum_client.estimate_data_gas(HexBytes("0x050204")),
            GAS_CALL_DATA_BYTE * 3,
        )
        self.assertEqual(
            self.ethereum_client.estimate_data_gas(HexBytes("0x0502040000")),
            GAS_CALL_DATA_BYTE * 3 + 4 * 2,
        )
        self.assertEqual(
            self.ethereum_client.estimate_data_gas(HexBytes("0x050204000001")),
            GAS_CALL_DATA_BYTE * 4 + 4 * 2,
        )
        self.assertEqual(
            self.ethereum_client.estimate_data_gas(HexBytes("0x00050204000001")),
            4 + GAS_CALL_DATA_BYTE * 4 + 4 * 2,
        )

    def test_provider_singleton(self):
        ethereum_client1 = EthereumClientProvider()
        ethereum_client2 = EthereumClientProvider()
        self.assertEqual(ethereum_client1, ethereum_client2)

    def test_send_eth_to(self):
        address, _ = get_eth_address_with_key()
        value = 1
        self.ethereum_client.send_eth_to(
            self.ethereum_test_account.key, address, self.gas_price, value
        )
        self.assertEqual(self.ethereum_client.get_balance(address), value)

    def test_send_eth_without_key(self):
        with self.settings(SAFE_FUNDER_PRIVATE_KEY=None):
            self.test_send_eth_to()

    def test_send_transaction(self):
        account = self.ethereum_test_account
        address = account.address
        to, _ = get_eth_address_with_key()
        value = 1
        tx = {
            "to": to,
            "value": value,
            "gas": 23000,
            "gasPrice": self.gas_price,
            "nonce": self.ethereum_client.get_nonce_for_account(address),
        }

        with self.assertRaises(FromAddressNotFound):
            self.ethereum_client.send_transaction(tx)

        # Random address not in the node
        tx["from"] = Account.create().address
        with self.assertRaises(SenderAccountNotFoundInNode):
            self.ethereum_client.send_transaction(tx)

        tx["from"] = address
        self.ethereum_client.send_transaction(tx)
        self.assertEqual(self.ethereum_client.get_balance(to), 1)

        with self.assertRaises(InvalidNonce):
            self.ethereum_client.send_transaction(tx)

        tx["value"] = self.ethereum_client.get_balance(address) + 1
        tx["nonce"] = self.ethereum_client.get_nonce_for_account(address)
        with self.assertRaises(InsufficientFunds):
            self.ethereum_client.send_transaction(tx)

    def test_send_unsigned_transaction(self):
        account = self.ethereum_test_account
        address = account.address
        to, _ = get_eth_address_with_key()
        value = 4

        tx = {
            "to": to,
            "value": value,
            "gas": 23000,
            "gasPrice": 1,
        }

        with self.assertRaisesMessage(
            ValueError, "Ethereum account was not configured"
        ):
            self.ethereum_client.send_unsigned_transaction(tx)

        self.ethereum_client.send_unsigned_transaction(tx, public_key=address)
        self.assertEqual(self.ethereum_client.get_balance(to), value)
        first_nonce = tx["nonce"]
        self.assertGreaterEqual(first_nonce, 0)

        # Will use the same nonce
        with self.assertRaisesMessage(InvalidNonce, "correct nonce"):
            self.ethereum_client.send_unsigned_transaction(tx, public_key=address)

        # With retry, everything should work
        self.ethereum_client.send_unsigned_transaction(
            tx, public_key=address, retry=True
        )
        self.assertEqual(tx["nonce"], first_nonce + 1)
        self.assertEqual(self.ethereum_client.get_balance(to), value * 2)

        # We try again with the first nonce, and should work too
        tx["nonce"] = first_nonce
        self.ethereum_client.send_unsigned_transaction(
            tx, public_key=address, retry=True
        )
        self.assertEqual(tx["nonce"], first_nonce + 2)
        self.assertEqual(self.ethereum_client.get_balance(to), value * 3)

    def test_send_unsigned_transaction_with_private_key(self):
        account = self.create_account(initial_ether=0.1)
        key = account.key
        to, _ = get_eth_address_with_key()
        value = 4

        tx = {
            "to": to,
            "value": value,
            "gas": 23000,
            "gasPrice": self.gas_price,
        }

        self.ethereum_client.send_unsigned_transaction(tx, private_key=key)
        self.assertEqual(self.ethereum_client.get_balance(to), value)
        first_nonce = tx["nonce"]
        self.assertGreaterEqual(first_nonce, 0)

        # Will use the same nonce
        with self.assertRaisesMessage(InvalidNonce, "correct nonce"):
            self.ethereum_client.send_unsigned_transaction(tx, private_key=key)

        # With retry, everything should work
        self.ethereum_client.send_unsigned_transaction(tx, private_key=key, retry=True)
        self.assertEqual(tx["nonce"], first_nonce + 1)
        self.assertEqual(self.ethereum_client.get_balance(to), value * 2)

        # We try again with the first nonce, and should work too
        tx["nonce"] = first_nonce
        self.ethereum_client.send_unsigned_transaction(tx, private_key=key, retry=True)
        self.assertEqual(tx["nonce"], first_nonce + 2)
        self.assertEqual(self.ethereum_client.get_balance(to), value * 3)

    def test_wait_for_tx_receipt(self):
        value = 1
        to = Account.create().address

        tx_hash = self.ethereum_client.send_eth_to(
            self.ethereum_test_account.key, to=to, gas_price=self.gas_price, value=value
        )
        receipt1 = self.ethereum_client.get_transaction_receipt(tx_hash, timeout=None)
        receipt2 = self.ethereum_client.get_transaction_receipt(tx_hash, timeout=20)
        self.assertIsNotNone(receipt1)
        self.assertEqual(receipt1, receipt2)

        fake_tx_hash = self.w3.keccak(0)
        self.assertIsNone(
            self.ethereum_client.get_transaction_receipt(fake_tx_hash, timeout=None)
        )
        self.assertIsNone(
            self.ethereum_client.get_transaction_receipt(fake_tx_hash, timeout=1)
        )

    def test_batch_call(self):
        account_address = Account.create().address
        tokens_value = 12
        tokens_value_2 = 25
        erc20 = self.deploy_example_erc20(tokens_value, account_address)
        erc20_2 = self.deploy_example_erc20(tokens_value_2, account_address)

        for batch_call_manager in (
            self.ethereum_client,
            self.ethereum_client.batch_call_manager,
        ):
            with self.subTest(batch_call_manager=batch_call_manager):
                results = batch_call_manager.batch_call(
                    [
                        erc20.functions.decimals(),
                        erc20.functions.symbol(),
                        erc20.functions.balanceOf(account_address),
                    ]
                )
                decimals, symbol, balance_of = results
                self.assertEqual(decimals, erc20.functions.decimals().call())
                self.assertEqual(symbol, erc20.functions.symbol().call())
                self.assertEqual(balance_of, tokens_value)

                results = batch_call_manager.batch_call_same_function(
                    erc20.functions.balanceOf(account_address),
                    [erc20.address, erc20_2.address],
                )
                self.assertEqual([tokens_value, tokens_value_2], results)

                invalid_erc20 = get_erc20_contract(
                    self.ethereum_client.w3, Account.create().address
                )
                with self.assertRaises(BatchCallException):
                    batch_call_manager.batch_call(
                        [
                            invalid_erc20.functions.decimals(),
                            invalid_erc20.functions.symbol(),
                            erc20.functions.balanceOf(account_address),
                            invalid_erc20.functions.balanceOf(account_address),
                        ]
                    )

                # It shouldn't raise error and instead return None
                self.assertEqual(
                    batch_call_manager.batch_call(
                        [
                            invalid_erc20.functions.decimals(),
                            invalid_erc20.functions.symbol(),
                            erc20.functions.balanceOf(account_address),
                            invalid_erc20.functions.balanceOf(account_address),
                        ],
                        raise_exception=False,
                    ),
                    [None, None, tokens_value, None],
                )

                with self.assertRaises(BatchCallException):
                    batch_call_manager.batch_call(
                        [
                            erc20.functions.decimals(),
                            erc20.functions.transfer(NULL_ADDRESS, 1),
                            erc20.functions.symbol(),
                            erc20.functions.balanceOf(account_address),
                        ],
                        raise_exception=True,
                    )

    def test_get_blocks(self):
        self.assertEqual(self.ethereum_client.get_blocks([]), [])
        # Generate 3 blocks
        to = Account.create().address
        value = 345
        for _ in range(3):
            self.send_ether(to, value)

        block_numbers = [
            self.ethereum_client.current_block_number - i for i in range(3)
        ]
        blocks = self.ethereum_client.get_blocks(block_numbers, full_transactions=True)
        block_hashes = [block["hash"] for block in blocks]
        block_hashes_hex = [block_hash.hex() for block_hash in block_hashes]
        for block_number, block in zip(block_numbers, blocks):
            self.assertEqual(block["number"], block_number)
            self.assertEqual(len(block["hash"]), 32)
            self.assertEqual(len(block["parentHash"]), 32)
            self.assertGreaterEqual(len(block["transactions"]), 0)

        blocks = self.ethereum_client.get_blocks(block_hashes, full_transactions=True)
        for block_number, block in zip(block_numbers, blocks):
            self.assertEqual(block["number"], block_number)
            self.assertEqual(len(block["hash"]), 32)
            self.assertEqual(len(block["parentHash"]), 32)
            self.assertGreaterEqual(len(block["transactions"]), 0)

        blocks = self.ethereum_client.get_blocks(
            block_hashes_hex, full_transactions=True
        )
        for block_number, block in zip(block_numbers, blocks):
            self.assertEqual(block["number"], block_number)
            self.assertEqual(len(block["hash"]), 32)
            self.assertEqual(len(block["parentHash"]), 32)
            self.assertGreaterEqual(len(block["transactions"]), 0)

    def test_is_eip1559_supported(self):
        self.assertFalse(self.ethereum_client.is_eip1559_supported())

    def test_set_eip1559_fees(self):
        with mock.patch.object(
            EthereumClient, "estimate_fee_eip1559", return_value=(2, 5)
        ):
            tx: TxParams = {"to": Account.create(), "value": 1, "gasPrice": 5}
            eip_1559_tx = self.ethereum_client.set_eip1559_fees(tx)
            self.assertIn("gasPrice", tx)  # Provided tx is not modified
            self.assertNotIn("gasPrice", eip_1559_tx)
            self.assertEqual(
                eip_1559_tx["chainId"], self.ethereum_client.get_network().value
            )
            self.assertEqual(eip_1559_tx["maxPriorityFeePerGas"], 5)
            self.assertEqual(eip_1559_tx["maxFeePerGas"], 7)


class TestEthereumClientWithMainnetNode(EthereumTestCaseMixin, TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        mainnet_node = just_test_if_mainnet_node()
        cls.ethereum_client = EthereumClient(mainnet_node)

    def test_is_eip1559_supported(self):
        self.assertTrue(self.ethereum_client.is_eip1559_supported())

    def test_estimate_fee_eip1559(self):
        """
        EIP1559 is still not supported on Ganache
        """
        (
            base_fee_per_gas,
            max_priority_fee_per_gas,
        ) = self.ethereum_client.estimate_fee_eip1559()
        self.assertGreater(base_fee_per_gas, 0)
        self.assertGreaterEqual(max_priority_fee_per_gas, 0)

    def test_trace_block(self):
        block_number = 2191709
        self.assertEqual(
            self.ethereum_client.parity.trace_block(block_number),
            trace_block_2191709_mock,
        )

    def test_trace_blocks(self):
        block_numbers = [13191781, 2191709]
        self.assertEqual(
            self.ethereum_client.parity.trace_blocks(block_numbers),
            [trace_block_13191781_mock, trace_block_2191709_mock],
        )

    def test_trace_transaction(self):
        tx_hash = "0x0b04589bdc11585fb98f270b1bfeff0fb3bbb3c56d35b104f62d8115d6f7c57f"  # Safe 1.3.0 deployment
        self.assertEqual(
            self.ethereum_client.parity.trace_transaction(tx_hash),
            trace_transaction_mocks[tx_hash],
        )

    def test_trace_transactions(self):
        tx_hashes = [
            "0x0b04589bdc11585fb98f270b1bfeff0fb3bbb3c56d35b104f62d8115d6f7c57f",  # Safe 1.3.0 deployment
            "0xf325b4e52d0649593e8c82f35bd389c13c13b21b61bc17de295979a21e5cfdc0",  # Safe 1.1.0 setup
        ]
        self.assertEqual(
            self.ethereum_client.parity.trace_transactions(tx_hashes),
            [trace_transaction_mocks[tx_hash] for tx_hash in tx_hashes],
        )

    def test_trace_filter(self):
        safe_1_3_0_address = "0xd9Db270c1B5E3Bd161E8c8503c55cEABeE709552"
        self.assertEqual(
            self.ethereum_client.parity.trace_filter(
                from_block=12504268, to_block=12504268, to_address=[safe_1_3_0_address]
            ),
            trace_filter_mock_1,
        )



"""
Support for MakerDAO MultiCall contract
"""
import logging
from dataclasses import dataclass
from typing import Any, List, Optional, Sequence, Tuple

from eth_abi.exceptions import DecodingError
from eth_account.signers.local import LocalAccount
from eth_typing import BlockIdentifier, BlockNumber, ChecksumAddress
from hexbytes import HexBytes
from web3 import Web3
from web3._utils.abi import map_abi_data
from web3._utils.normalizers import BASE_RETURN_NORMALIZERS
from web3.contract import ContractFunction
from web3.exceptions import ContractLogicError

from . import EthereumClient, EthereumNetwork, EthereumNetworkNotSupported
from .ethereum_client import EthereumTxSent
from .exceptions import BatchCallFunctionFailed
from .oracles.abis.makerdao import multicall_v2_abi, multicall_v2_bytecode

logger = logging.getLogger(__name__)


@dataclass
class MulticallResult:
    success: bool
    return_data: Optional[bytes]


@dataclass
class MulticallDecodedResult:
    success: bool
    return_data_decoded: Optional[Any]


class Multicall:
    ADDRESSES = {
        EthereumNetwork.MAINNET: "0x5BA1e12693Dc8F9c48aAD8770482f4739bEeD696",
        EthereumNetwork.ARBITRUM: "0x021CeAC7e681dBCE9b5039d2535ED97590eB395c",
        EthereumNetwork.BINANCE: "0xed386Fe855C1EFf2f843B910923Dd8846E45C5A4",
        EthereumNetwork.GOERLI: "0x5BA1e12693Dc8F9c48aAD8770482f4739bEeD696",
        EthereumNetwork.KOVAN: "0x5BA1e12693Dc8F9c48aAD8770482f4739bEeD696",
        EthereumNetwork.MATIC: "0xed386Fe855C1EFf2f843B910923Dd8846E45C5A4",
        EthereumNetwork.MUMBAI: "0xed386Fe855C1EFf2f843B910923Dd8846E45C5A4",
        EthereumNetwork.RINKEBY: "0x5BA1e12693Dc8F9c48aAD8770482f4739bEeD696",
        EthereumNetwork.ROPSTEN: "0x5BA1e12693Dc8F9c48aAD8770482f4739bEeD696",
        EthereumNetwork.XDAI: "0x08612d3C4A5Dfe2FaaFaFe6a4ff712C2dC675bF7",
        EthereumNetwork.FANTOM: "0xD98e3dBE5950Ca8Ce5a4b59630a5652110403E5c",
        EthereumNetwork.AVALANCHE: "0xAbeC56f92a89eEe33F5194Ca4151DD59785c2C74",
    }

    def __init__(
        self,
        ethereum_client: EthereumClient,
        multicall_contract_address: Optional[ChecksumAddress] = None,
    ):
        self.ethereum_client = ethereum_client
        self.w3 = ethereum_client.w3
        ethereum_network = ethereum_client.get_network()
        address = multicall_contract_address or self.ADDRESSES.get(ethereum_network)
        if not address:
            raise EthereumNetworkNotSupported(
                "Multicall contract not available for %s", ethereum_network.name
            )
        self.contract = self.get_contract(self.w3, address)

    def get_contract(self, w3: Web3, address: Optional[ChecksumAddress] = None):
        return w3.eth.contract(
            address, abi=multicall_v2_abi, bytecode=multicall_v2_bytecode
        )

    @classmethod
    def deploy_contract(
        cls, ethereum_client: EthereumClient, deployer_account: LocalAccount
    ) -> EthereumTxSent:
        """
        Deploy contract

        :param ethereum_client:
        :param deployer_account: Ethereum Account
        :return: deployed contract address
        """
        contract = cls.get_contract(cls, ethereum_client.w3)
        tx = contract.constructor().buildTransaction({"from": deployer_account.address})

        tx_hash = ethereum_client.send_unsigned_transaction(
            tx, private_key=deployer_account.key
        )
        tx_receipt = ethereum_client.get_transaction_receipt(tx_hash, timeout=120)
        assert tx_receipt and tx_receipt["status"]
        contract_address = tx_receipt["contractAddress"]
        logger.info(
            "Deployed Multicall V2 Contract %s by %s",
            contract_address,
            deployer_account.address,
        )
        # Add address to addresses dictionary
        cls.ADDRESSES[ethereum_client.get_network()] = contract_address
        return EthereumTxSent(tx_hash, tx, contract_address)

    @staticmethod
    def _build_payload(
        contract_functions: Sequence[ContractFunction],
    ) -> Tuple[List[Tuple[ChecksumAddress, bytes]], List[List[Any]]]:
        targets_with_data = []
        output_types = []
        for contract_function in contract_functions:
            targets_with_data.append(
                (
                    contract_function.address,
                    HexBytes(contract_function._encode_transaction_data()),
                )
            )
            output_types.append(
                [output["type"] for output in contract_function.abi["outputs"]]
            )

        return targets_with_data, output_types

    def _build_payload_same_function(
        self,
        contract_function: ContractFunction,
        contract_addresses: Sequence[ChecksumAddress],
    ) -> Tuple[List[Tuple[ChecksumAddress, bytes]], List[List[Any]]]:
        targets_with_data = []
        output_types = []
        tx_data = HexBytes(contract_function._encode_transaction_data())
        for contract_address in contract_addresses:
            targets_with_data.append((contract_address, tx_data))
            output_types.append(
                [output["type"] for output in contract_function.abi["outputs"]]
            )

        return targets_with_data, output_types

    def _decode_data(self, output_type: Sequence[str], data: bytes) -> Optional[Any]:
        """

        :param output_type:
        :param data:
        :return:
        :raises: DecodingError
        """
        if data:
            try:
                decoded_values = self.w3.codec.decode_abi(output_type, data)
                normalized_data = map_abi_data(
                    BASE_RETURN_NORMALIZERS, output_type, decoded_values
                )
                if len(normalized_data) == 1:
                    return normalized_data[0]
                else:
                    return normalized_data
            except DecodingError:
                logger.warning(
                    "Cannot decode %s using output-type %s", data, output_type
                )
                return data

    def _aggregate(
        self,
        targets_with_data: Sequence[Tuple[ChecksumAddress, bytes]],
        block_identifier: Optional[BlockIdentifier] = "latest",
    ) -> Tuple[BlockNumber, List[Optional[Any]]]:
        """

        :param targets_with_data: List of target `addresses` and `data` to be called in each Contract
        :param block_identifier:
        :return:
        :raises: BatchCallFunctionFailed
        """
        aggregate_parameter = [
            {"target": target, "callData": data} for target, data in targets_with_data
        ]
        try:
            return self.contract.functions.aggregate(aggregate_parameter).call(
                block_identifier=block_identifier
            )
        except (ContractLogicError, OverflowError):
            raise BatchCallFunctionFailed

    def aggregate(
        self,
        contract_functions: Sequence[ContractFunction],
        block_identifier: Optional[BlockIdentifier] = "latest",
    ) -> Tuple[BlockNumber, List[Optional[Any]]]:
        """
        Calls ``aggregate`` on MakerDAO's Multicall contract. If a function called raises an error execution is stopped

        :param contract_functions:
        :param block_identifier:
        :return: A tuple with the ``blockNumber`` and a list with the decoded return values
        :raises: BatchCallFunctionFailed
        """
        targets_with_data, output_types = self._build_payload(contract_functions)
        block_number, results = self._aggregate(
            targets_with_data, block_identifier=block_identifier
        )
        decoded_results = [
            self._decode_data(output_type, data)
            for output_type, data in zip(output_types, results)
        ]
        return block_number, decoded_results

    def _try_aggregate(
        self,
        targets_with_data: Sequence[Tuple[ChecksumAddress, bytes]],
        require_success: bool = False,
        block_identifier: Optional[BlockIdentifier] = "latest",
    ) -> List[MulticallResult]:
        """
        Calls ``try_aggregate`` on MakerDAO's Multicall contract.

        :param targets_with_data:
        :param require_success: If ``True``, an exception in any of the functions will stop the execution. Also, an
            invalid decoded value will stop the execution
        :param block_identifier:
        :return: A list with the decoded return values
        """

        aggregate_parameter = [
            {"target": target, "callData": data} for target, data in targets_with_data
        ]
        try:
            result = self.contract.functions.tryAggregate(
                require_success, aggregate_parameter
            ).call(block_identifier=block_identifier)

            if require_success and b"" in (data for _, data in result):
                # `b''` values are decoding errors/missing contracts/missing functions
                raise BatchCallFunctionFailed

            return [
                MulticallResult(success, data if data else None)
                for success, data in result
            ]
        except (ContractLogicError, OverflowError):
            raise BatchCallFunctionFailed

    def try_aggregate(
        self,
        contract_functions: Sequence[ContractFunction],
        require_success: bool = False,
        block_identifier: Optional[BlockIdentifier] = "latest",
    ) -> List[MulticallDecodedResult]:
        """
        Calls ``try_aggregate`` on MakerDAO's Multicall contract.

        :param contract_functions:
        :param require_success: If ``True``, an exception in any of the functions will stop the execution
        :param block_identifier:
        :return: A list with the decoded return values
        """
        targets_with_data, output_types = self._build_payload(contract_functions)
        results = self._try_aggregate(
            targets_with_data,
            require_success=require_success,
            block_identifier=block_identifier,
        )
        return [
            MulticallDecodedResult(
                multicall_result.success,
                self._decode_data(output_type, multicall_result.return_data)
                if multicall_result.success
                else multicall_result.return_data,
            )
            for output_type, multicall_result in zip(output_types, results)
        ]

    def try_aggregate_same_function(
        self,
        contract_function: ContractFunction,
        contract_addresses: Sequence[ChecksumAddress],
        require_success: bool = False,
        block_identifier: Optional[BlockIdentifier] = "latest",
    ) -> List[MulticallDecodedResult]:
        """
        Calls ``try_aggregate`` on MakerDAO's Multicall contract. Reuse same function with multiple contract addresses.
        It's more optimal due to instantiating ``ContractFunction`` objects is very demanding

        :param contract_function:
        :param contract_addresses:
        :param require_success: If ``True``, an exception in any of the functions will stop the execution
        :param block_identifier:
        :return: A list with the decoded return values
        """

        targets_with_data, output_types = self._build_payload_same_function(
            contract_function, contract_addresses
        )
        results = self._try_aggregate(
            targets_with_data,
            require_success=require_success,
            block_identifier=block_identifier,
        )
        return [
            MulticallDecodedResult(
                multicall_result.success,
                self._decode_data(output_type, multicall_result.return_data)
                if multicall_result.success
                else multicall_result.return_data,
            )
            for output_type, multicall_result in zip(output_types, results)
        ]



import logging
from typing import List, Optional

from django.conf import settings

from eth_account import Account
from eth_typing import ChecksumAddress
from hexbytes import HexBytes
from web3.contract import Contract

from gnosis.eth.contracts import (
    get_compatibility_fallback_handler_V1_3_0_contract,
    get_multi_send_contract,
    get_proxy_factory_contract,
    get_safe_V1_0_0_contract,
    get_safe_V1_1_1_contract,
    get_safe_V1_3_0_contract,
)
from gnosis.eth.tests.ethereum_test_case import EthereumTestCaseMixin
from gnosis.safe import Safe
from gnosis.safe.multi_send import MultiSend
from gnosis.safe.proxy_factory import ProxyFactory
from gnosis.safe.safe_create2_tx import SafeCreate2Tx

from ...eth.constants import NULL_ADDRESS
from .utils import generate_salt_nonce

logger = logging.getLogger(__name__)


_contract_addresses = {
    "safe_V0_0_1": Safe.deploy_master_contract_v0_0_1,
    "safe_V1_0_0": Safe.deploy_master_contract_v1_0_0,
    "safe_V1_1_1": Safe.deploy_master_contract_v1_1_1,
    "safe_v1_3_0": Safe.deploy_master_contract_v1_3_0,
    "compatibility_fallback_handler": Safe.deploy_compatibility_fallback_handler,
    "proxy_factory": ProxyFactory.deploy_proxy_factory_contract,
    "proxy_factory_V1_0_0": ProxyFactory.deploy_proxy_factory_contract_v1_0_0,
    "multi_send": MultiSend.deploy_contract,
}


class SafeTestCaseMixin(EthereumTestCaseMixin):
    safe_contract_address: ChecksumAddress
    safe_contract: Contract
    safe_contract_V1_1_1_address: ChecksumAddress
    safe_contract_V1_1_1: Contract
    safe_contract_V1_0_0_address: ChecksumAddress
    safe_contract_V1_0_0: Contract
    safe_contract_V0_0_1_address: ChecksumAddress
    safe_contract_V0_0_1: Contract
    proxy_factory_contract_address: ChecksumAddress
    proxy_factory_contract: Contract
    proxy_factory: ProxyFactory
    multi_send_contract: Contract
    multi_send: MultiSend

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        for key, value in _contract_addresses.items():
            if callable(value):
                _contract_addresses[key] = value(
                    cls.ethereum_client, cls.ethereum_test_account
                ).contract_address

        settings.SAFE_DEFAULT_CALLBACK_HANDLER = _contract_addresses[
            "compatibility_fallback_handler"
        ]
        settings.SAFE_MULTISEND_ADDRESS = _contract_addresses["multi_send"]
        settings.SAFE_CONTRACT_ADDRESS = _contract_addresses["safe_v1_3_0"]
        settings.SAFE_V1_1_1_CONTRACT_ADDRESS = _contract_addresses["safe_V1_1_1"]
        settings.SAFE_V1_0_0_CONTRACT_ADDRESS = _contract_addresses["safe_V1_0_0"]
        settings.SAFE_V0_0_1_CONTRACT_ADDRESS = _contract_addresses["safe_V0_0_1"]
        settings.SAFE_PROXY_FACTORY_ADDRESS = _contract_addresses["proxy_factory"]
        settings.SAFE_PROXY_FACTORY_V1_0_0_ADDRESS = _contract_addresses[
            "proxy_factory_V1_0_0"
        ]
        settings.SAFE_VALID_CONTRACT_ADDRESSES = {
            settings.SAFE_CONTRACT_ADDRESS,
            settings.SAFE_V1_1_1_CONTRACT_ADDRESS,
            settings.SAFE_V1_0_0_CONTRACT_ADDRESS,
            settings.SAFE_V0_0_1_CONTRACT_ADDRESS,
        }
        cls.compatibility_fallback_handler = (
            get_compatibility_fallback_handler_V1_3_0_contract(
                cls.w3, _contract_addresses["compatibility_fallback_handler"]
            )
        )
        cls.safe_contract_address = _contract_addresses["safe_v1_3_0"]
        cls.safe_contract = get_safe_V1_3_0_contract(cls.w3, cls.safe_contract_address)
        cls.safe_contract_V1_1_1_address = _contract_addresses["safe_V1_1_1"]
        cls.safe_contract_V1_1_1 = get_safe_V1_1_1_contract(
            cls.w3, cls.safe_contract_V1_1_1_address
        )
        cls.safe_contract_V1_0_0_address = _contract_addresses["safe_V1_0_0"]
        cls.safe_contract_V1_0_0 = get_safe_V1_0_0_contract(
            cls.w3, cls.safe_contract_V1_0_0_address
        )
        cls.safe_contract_V0_0_1_address = _contract_addresses["safe_V0_0_1"]
        cls.safe_contract_V0_0_1 = get_safe_V1_0_0_contract(
            cls.w3, cls.safe_contract_V0_0_1_address
        )
        cls.proxy_factory_contract_address = _contract_addresses["proxy_factory"]
        cls.proxy_factory_contract = get_proxy_factory_contract(
            cls.w3, cls.proxy_factory_contract_address
        )
        cls.proxy_factory = ProxyFactory(
            cls.proxy_factory_contract_address, cls.ethereum_client
        )
        cls.multi_send_contract = get_multi_send_contract(
            cls.w3, _contract_addresses["multi_send"]
        )
        cls.multi_send = MultiSend(cls.multi_send_contract.address, cls.ethereum_client)

    def build_test_safe(
        self,
        number_owners: int = 3,
        threshold: Optional[int] = None,
        owners: Optional[List[str]] = None,
        fallback_handler: Optional[str] = None,
    ) -> SafeCreate2Tx:
        salt_nonce = generate_salt_nonce()
        owners = (
            owners
            if owners
            else [Account.create().address for _ in range(number_owners)]
        )
        threshold = threshold if threshold else len(owners) - 1

        gas_price = self.ethereum_client.w3.eth.gas_price
        return Safe.build_safe_create2_tx(
            self.ethereum_client,
            self.safe_contract_address,
            self.proxy_factory_contract_address,
            salt_nonce,
            owners,
            threshold,
            fallback_handler=fallback_handler,
            gas_price=gas_price,
            payment_token=None,
            fixed_creation_cost=0,
        )

    def deploy_test_safe(
        self,
        number_owners: int = 3,
        threshold: Optional[int] = None,
        owners: Optional[List[ChecksumAddress]] = None,
        initial_funding_wei: int = 0,
        fallback_handler: ChecksumAddress = None,
    ) -> Safe:
        """
        Deploy a Safe v1.3.0

        :param number_owners:
        :param threshold:
        :param owners:
        :param initial_funding_wei:
        :param fallback_handler:
        :return:
        """
        fallback_handler = (
            fallback_handler or self.compatibility_fallback_handler.address
        )
        owners = (
            owners
            if owners
            else [Account.create().address for _ in range(number_owners)]
        )
        if not threshold:
            threshold = len(owners) - 1 if len(owners) > 1 else 1
        empty_parameters = {"gas": 1, "gasPrice": 1}
        to = NULL_ADDRESS
        data = b""
        payment_token = NULL_ADDRESS
        payment = 0
        payment_receiver = NULL_ADDRESS
        initializer = HexBytes(
            self.safe_contract.functions.setup(
                owners,
                threshold,
                to,
                data,
                fallback_handler,
                payment_token,
                payment,
                payment_receiver,
            ).buildTransaction(empty_parameters)["data"]
        )
        ethereum_tx_sent = self.proxy_factory.deploy_proxy_contract(
            self.ethereum_test_account,
            self.safe_contract.address,
            initializer=initializer,
        )
        safe = Safe(ethereum_tx_sent.contract_address, self.ethereum_client)
        if initial_funding_wei:
            self.send_ether(safe.address, initial_funding_wei)

        self.assertEqual(safe.retrieve_version(), "1.3.0")
        self.assertEqual(safe.retrieve_threshold(), threshold)
        self.assertCountEqual(safe.retrieve_owners(), owners)

        return safe

    def deploy_test_safe_v1_1_1(
        self,
        number_owners: int = 3,
        threshold: Optional[int] = None,
        owners: Optional[List[str]] = None,
        initial_funding_wei: int = 0,
        fallback_handler: Optional[str] = NULL_ADDRESS,
    ) -> Safe:
        owners = (
            owners
            if owners
            else [Account.create().address for _ in range(number_owners)]
        )
        if not threshold:
            threshold = len(owners) - 1 if len(owners) > 1 else 1
        empty_parameters = {"gas": 1, "gasPrice": 1}
        to = NULL_ADDRESS
        data = b""
        payment_token = NULL_ADDRESS
        payment = 0
        payment_receiver = NULL_ADDRESS
        initializer = HexBytes(
            self.safe_contract_V1_1_1.functions.setup(
                owners,
                threshold,
                to,
                data,
                fallback_handler,
                payment_token,
                payment,
                payment_receiver,
            ).buildTransaction(empty_parameters)["data"]
        )
        ethereum_tx_sent = self.proxy_factory.deploy_proxy_contract(
            self.ethereum_test_account,
            self.safe_contract_V1_1_1.address,
            initializer=initializer,
        )
        safe = Safe(ethereum_tx_sent.contract_address, self.ethereum_client)
        if initial_funding_wei:
            self.send_ether(safe.address, initial_funding_wei)

        self.assertEqual(safe.retrieve_version(), "1.1.1")
        self.assertEqual(safe.retrieve_threshold(), threshold)
        self.assertCountEqual(safe.retrieve_owners(), owners)

        return safe

    def deploy_test_safe_v1_0_0(
        self,
        number_owners: int = 3,
        threshold: Optional[int] = None,
        owners: Optional[List[ChecksumAddress]] = None,
        initial_funding_wei: int = 0,
    ) -> Safe:
        owners = (
            owners
            if owners
            else [Account.create().address for _ in range(number_owners)]
        )
        if not threshold:
            threshold = len(owners) - 1 if len(owners) > 1 else 1
        empty_parameters = {"gas": 1, "gasPrice": 1}
        to = NULL_ADDRESS
        data = b""
        payment_token = NULL_ADDRESS
        payment = 0
        payment_receiver = NULL_ADDRESS
        initializer = HexBytes(
            self.safe_contract_V1_0_0.functions.setup(
                owners, threshold, to, data, payment_token, payment, payment_receiver
            ).buildTransaction(empty_parameters)["data"]
        )
        ethereum_tx_sent = self.proxy_factory.deploy_proxy_contract(
            self.ethereum_test_account,
            self.safe_contract_V1_0_0_address,
            initializer=initializer,
        )
        safe = Safe(ethereum_tx_sent.contract_address, self.ethereum_client)
        if initial_funding_wei:
            self.send_ether(safe.address, initial_funding_wei)

        self.assertEqual(safe.retrieve_version(), "1.0.0")
        self.assertEqual(safe.retrieve_threshold(), threshold)
        self.assertCountEqual(safe.retrieve_owners(), owners)

        return safe



from unittest import mock
from unittest.mock import MagicMock

from django.test import TestCase

from .. import EthereumClient, EthereumNetwork, EthereumNetworkNotSupported
from ..constants import NULL_ADDRESS
from ..contracts import get_erc20_contract
from ..exceptions import BatchCallFunctionFailed
from ..multicall import Multicall, MulticallDecodedResult
from .ethereum_test_case import EthereumTestCaseMixin
from .utils import just_test_if_mainnet_node


class TestMulticallGanache(EthereumTestCaseMixin, TestCase):
    """
    Test Multicall using ganache
    """

    @mock.patch.object(
        EthereumClient, "get_network", return_value=EthereumNetwork.UNKNOWN
    )
    def test_multicall_network_exception(self, ethereum_network_mock: MagicMock):
        with self.assertRaisesMessage(
            EthereumNetworkNotSupported, "Multicall contract not available for"
        ):
            Multicall(self.ethereum_client)

    def test_multicall_try_aggregate(self):
        erc20_contract = self.deploy_example_erc20(
            1, self.ethereum_test_account.address
        )

        results = self.multicall.try_aggregate(
            [
                erc20_contract.functions.name(),
                erc20_contract.functions.transfer(NULL_ADDRESS, 1),
                erc20_contract.functions.symbol(),
                erc20_contract.functions.decimals(),
            ]
        )
        expected_results = [
            MulticallDecodedResult(success=True, return_data_decoded="Uxio"),
            MulticallDecodedResult(
                success=False,
                return_data_decoded=b"\x08\xc3y\xa0\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00 \x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00#ERC20: transfer to the zero address\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00",
            ),  # noqa
            MulticallDecodedResult(success=True, return_data_decoded="UXI"),
            MulticallDecodedResult(success=True, return_data_decoded=18),
        ]

        self.assertEqual(results, expected_results)

    def test_try_aggregate_same_function(self):
        account = "0x44425Eff435c3B1C38Cbc1139f248CA2918ACccA"
        erc20_contracts = [self.deploy_example_erc20(i, account) for i in range(3)]
        token_addresses = [erc20_contract.address for erc20_contract in erc20_contracts]

        erc20_contract = get_erc20_contract(self.ethereum_client.w3)

        results = self.multicall.try_aggregate_same_function(
            erc20_contract.functions.balanceOf(account), token_addresses
        )
        expected_results = [
            MulticallDecodedResult(success=True, return_data_decoded=0),
            MulticallDecodedResult(success=True, return_data_decoded=1),
            MulticallDecodedResult(success=True, return_data_decoded=2),
        ]

        self.assertEqual(results, expected_results)


class TestMulticallNode(EthereumTestCaseMixin, TestCase):
    """
    Test Multicall using a production node
    """

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        mainnet_node = just_test_if_mainnet_node()
        cls.ethereum_client = EthereumClient(mainnet_node)
        cls.gno_contract = get_erc20_contract(
            cls.ethereum_client.w3, "0x6810e776880C02933D47DB1b9fc05908e5386b96"
        )
        cls.not_existing_contract = get_erc20_contract(
            cls.ethereum_client.w3, NULL_ADDRESS
        )
        cls.multicall = Multicall(cls.ethereum_client)

    def test_multicall_aggregate(self):
        block_number, results = self.multicall.aggregate(
            [
                self.gno_contract.functions.name(),
                self.gno_contract.functions.symbol(),
                self.gno_contract.functions.decimals(),
            ]
        )
        expected_results = ["Gnosis Token", "GNO", 18]
        self.assertEqual(results, expected_results)
        self.assertGreater(block_number, 0)

        expected_results = [None]
        block_number, results = self.multicall.aggregate(
            [
                self.not_existing_contract.functions.name(),
            ]
        )
        self.assertEqual(results, expected_results)
        self.assertGreater(block_number, 0)

        expected_results = ["Gnosis Token", None, "GNO", 18]
        block_number, results = self.multicall.aggregate(
            [
                self.gno_contract.functions.name(),
                self.not_existing_contract.functions.name(),
                self.gno_contract.functions.symbol(),
                self.gno_contract.functions.decimals(),
            ]
        )
        self.assertEqual(results, expected_results)
        self.assertGreater(block_number, 0)

        with self.assertRaises(BatchCallFunctionFailed):
            self.multicall.aggregate(
                [
                    self.gno_contract.functions.name(),
                    self.gno_contract.functions.transfer(NULL_ADDRESS, 1),
                    self.gno_contract.functions.symbol(),
                    self.gno_contract.functions.decimals(),
                ]
            )

    def test_multicall_ty_aggregate(self):
        results = self.multicall.try_aggregate(
            [
                self.gno_contract.functions.name(),
                self.gno_contract.functions.symbol(),
                self.gno_contract.functions.decimals(),
            ]
        )
        expected_results = [
            MulticallDecodedResult(success=True, return_data_decoded="Gnosis Token"),
            MulticallDecodedResult(success=True, return_data_decoded="GNO"),
            MulticallDecodedResult(success=True, return_data_decoded=18),
        ]
        self.assertEqual(results, expected_results)

        results = self.multicall.try_aggregate(
            [
                self.gno_contract.functions.name(),
                self.gno_contract.functions.transfer(NULL_ADDRESS, 1),
                self.gno_contract.functions.symbol(),
                self.gno_contract.functions.decimals(),
            ]
        )
        expected_results = [
            MulticallDecodedResult(success=True, return_data_decoded="Gnosis Token"),
            MulticallDecodedResult(success=False, return_data_decoded=None),
            MulticallDecodedResult(success=True, return_data_decoded="GNO"),
            MulticallDecodedResult(success=True, return_data_decoded=18),
        ]
        self.assertEqual(results, expected_results)

        with self.assertRaises(BatchCallFunctionFailed):
            self.multicall.try_aggregate(
                [
                    self.gno_contract.functions.name(),
                    self.gno_contract.functions.transfer(NULL_ADDRESS, 1),
                    self.gno_contract.functions.symbol(),
                    self.gno_contract.functions.decimals(),
                ],
                require_success=True,
            )



from abc import ABC, abstractmethod
from enum import Enum
from logging import getLogger
from typing import List, Union

from eth_abi import decode_single, encode_single
from eth_abi.exceptions import DecodingError
from eth_account.messages import defunct_hash_message
from eth_utils import to_checksum_address
from hexbytes import HexBytes
from web3.exceptions import BadFunctionCallOutput

from gnosis.eth import EthereumClient
from gnosis.eth.contracts import get_safe_contract, get_safe_V1_1_1_contract
from gnosis.safe.signatures import (
    get_signing_address,
    signature_split,
    signature_to_bytes,
)

logger = getLogger(__name__)


EthereumBytes = Union[bytes, str]


class SafeSignatureException(Exception):
    pass


class CannotCheckEIP1271ContractSignature(SafeSignatureException):
    pass


class SafeSignatureType(Enum):
    CONTRACT_SIGNATURE = 0
    APPROVED_HASH = 1
    EOA = 2
    ETH_SIGN = 3

    @staticmethod
    def from_v(v: int):
        if v == 0:
            return SafeSignatureType.CONTRACT_SIGNATURE
        elif v == 1:
            return SafeSignatureType.APPROVED_HASH
        elif v > 30:
            return SafeSignatureType.ETH_SIGN
        else:
            return SafeSignatureType.EOA


class SafeSignature(ABC):
    def __init__(self, signature: EthereumBytes, safe_tx_hash: EthereumBytes):
        self.signature = HexBytes(signature)
        self.safe_tx_hash = HexBytes(safe_tx_hash)
        self.v, self.r, self.s = signature_split(self.signature)

    def __str__(self):
        return f"SafeSignature type={self.signature_type.name} owner={self.owner}"

    @classmethod
    def parse_signature(
        cls,
        signatures: EthereumBytes,
        safe_tx_hash: EthereumBytes,
        ignore_trailing: bool = True,
    ) -> List["SafeSignature"]:
        """
        :param signatures: One or more signatures appended. EIP1271 data at the end is supported.
        :param safe_tx_hash:
        :param ignore_trailing: Ignore trailing data on the signature. Some libraries pad it and add some zeroes at
            the end
        :return: List of SafeSignatures decoded
        """
        if not signatures:
            return []
        elif isinstance(signatures, str):
            signatures = HexBytes(signatures)

        signature_size = 65  # For contract signatures there'll be some data at the end
        data_position = len(
            signatures
        )  # For contract signatures, to stop parsing at data position

        safe_signatures = []
        for i in range(0, len(signatures), signature_size):
            if (
                i >= data_position
            ):  # If contract signature data position is reached, stop
                break

            signature = signatures[i : i + signature_size]
            if ignore_trailing and len(signature) < 65:
                # Trailing stuff
                break
            v, r, s = signature_split(signature)
            signature_type = SafeSignatureType.from_v(v)
            safe_signature: "SafeSignature"
            if signature_type == SafeSignatureType.CONTRACT_SIGNATURE:
                if s < data_position:
                    data_position = s
                contract_signature_len = int.from_bytes(
                    signatures[s : s + 32], "big"
                )  # Len size is 32 bytes
                contract_signature = signatures[
                    s + 32 : s + 32 + contract_signature_len
                ]  # Skip array size (32 bytes)
                safe_signature = SafeSignatureContract(
                    signature, safe_tx_hash, contract_signature
                )
            elif signature_type == SafeSignatureType.APPROVED_HASH:
                safe_signature = SafeSignatureApprovedHash(signature, safe_tx_hash)
            elif signature_type == SafeSignatureType.EOA:
                safe_signature = SafeSignatureEOA(signature, safe_tx_hash)
            elif signature_type == SafeSignatureType.ETH_SIGN:
                safe_signature = SafeSignatureEthSign(signature, safe_tx_hash)

            safe_signatures.append(safe_signature)
        return safe_signatures

    def export_signature(self) -> HexBytes:
        """
        Exports signature in a format that's valid individually. That's important for contract signatures, as it
        will fix the offset
        :return:
        """
        return self.signature

    @property
    @abstractmethod
    def owner(self):
        """
        :return: Decode owner from signature, without any further validation (signature can be not valid)
        """
        raise NotImplementedError

    @abstractmethod
    def is_valid(self, ethereum_client: EthereumClient, safe_address: str) -> bool:
        """
        :param ethereum_client: Required for Contract Signature and Approved Hash check
        :param safe_address: Required for Approved Hash check
        :return: `True` if signature is valid, `False` otherwise
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def signature_type(self) -> SafeSignatureType:
        raise NotImplementedError


class SafeSignatureContract(SafeSignature):
    EIP1271_MAGIC_VALUE = HexBytes(0x20C13B0B)
    EIP1271_MAGIC_VALUE_UPDATED = HexBytes(0x1626BA7E)

    def __init__(
        self,
        signature: EthereumBytes,
        safe_tx_hash: EthereumBytes,
        contract_signature: EthereumBytes,
    ):
        super().__init__(signature, safe_tx_hash)
        self.contract_signature = HexBytes(contract_signature)

    @property
    def owner(self):
        """
        :return: Address of contract signing. No further checks to get the owner are needed,
            but it could be a non existing contract
        """
        return to_checksum_address(
            decode_single("address", encode_single("uint", self.r))
        )

    @property
    def signature_type(self):
        return SafeSignatureType.CONTRACT_SIGNATURE

    def export_signature(self) -> HexBytes:
        """
        Fix offset (s) and append `contract_signature` at the end of the signature
        :return:
        """
        return HexBytes(
            signature_to_bytes(self.v, self.r, 65)
            + encode_single("bytes", self.contract_signature)
        )

    def is_valid(self, ethereum_client: EthereumClient, *args) -> bool:
        safe_contract = get_safe_V1_1_1_contract(ethereum_client.w3, self.owner)
        # Newest versions of the Safe contract have `isValidSignature` on the compatibility fallback handler
        for block_identifier in ("pending", "latest"):
            try:
                return safe_contract.functions.isValidSignature(
                    self.safe_tx_hash, self.contract_signature
                ).call(block_identifier=block_identifier) in (
                    self.EIP1271_MAGIC_VALUE,
                    self.EIP1271_MAGIC_VALUE_UPDATED,
                )
            except (ValueError, BadFunctionCallOutput, DecodingError):
                # Error using `pending` block identifier or contract does not exist
                logger.warning(
                    "Cannot check EIP1271 signature from contract %s", self.owner
                )
        return False


class SafeSignatureApprovedHash(SafeSignature):
    @property
    def owner(self):
        return to_checksum_address(
            decode_single("address", encode_single("uint", self.r))
        )

    @property
    def signature_type(self):
        return SafeSignatureType.APPROVED_HASH

    @classmethod
    def build_for_owner(
        cls, owner: str, safe_tx_hash: str
    ) -> "SafeSignatureApprovedHash":
        r = owner.lower().replace("0x", "").rjust(64, "0")
        s = "0" * 64
        v = "01"
        return cls(HexBytes(r + s + v), safe_tx_hash)

    def is_valid(self, ethereum_client: EthereumClient, safe_address: str) -> bool:
        safe_contract = get_safe_contract(ethereum_client.w3, safe_address)
        exception: Exception
        for block_identifier in ("pending", "latest"):
            try:
                return (
                    safe_contract.functions.approvedHashes(
                        self.owner, self.safe_tx_hash
                    ).call(block_identifier=block_identifier)
                    == 1
                )
            except BadFunctionCallOutput as e:  # Error using `pending` block identifier
                exception = e
        raise exception  # This should never happen


class SafeSignatureEthSign(SafeSignature):
    @property
    def owner(self):
        # defunct_hash_message prepends `\x19Ethereum Signed Message:\n32`
        message_hash = defunct_hash_message(primitive=self.safe_tx_hash)
        return get_signing_address(message_hash, self.v - 4, self.r, self.s)

    @property
    def signature_type(self):
        return SafeSignatureType.ETH_SIGN

    def is_valid(self, *args) -> bool:
        return True


class SafeSignatureEOA(SafeSignature):
    @property
    def owner(self):
        return get_signing_address(self.safe_tx_hash, self.v, self.r, self.s)

    @property
    def signature_type(self):
        return SafeSignatureType.EOA

    def is_valid(self, *args) -> bool:
        return True



import binascii

from django.core.exceptions import ValidationError
from django.forms import CharField as CharFieldForm
from django.utils.translation import gettext_lazy as _

import django_filters
from hexbytes import HexBytes
from web3 import Web3


class EthereumAddressFieldForm(CharFieldForm):
    default_error_messages = {
        "invalid": _("Enter a valid checksummed Ethereum Address."),
    }

    def prepare_value(self, value):
        return value

    def to_python(self, value):
        value = super().to_python(value)
        if value in self.empty_values:
            return None
        elif not Web3.isChecksumAddress(value):
            raise ValidationError(self.error_messages["invalid"], code="invalid")
        return value


class EthereumAddressFilter(django_filters.Filter):
    field_class = EthereumAddressFieldForm


class Keccak256FieldForm(CharFieldForm):
    default_error_messages = {
        "invalid": _('"%(value)s" is not a valid keccak256 hash.'),
        "length": _('"%(value)s" keccak256 hash should be 32 bytes.'),
    }

    def prepare_value(self, value):
        return value

    def to_python(self, value):
        value = super().to_python(value)
        if value in self.empty_values:
            return None
        else:
            try:
                bytes_value = HexBytes(value)
                if len(bytes_value) != 32:
                    raise ValidationError(
                        self.error_messages["length"],
                        code="length",
                        params={"value": value},
                    )
            except (binascii.Error, ValueError):
                raise ValidationError(
                    self.error_messages["invalid"],
                    code="invalid",
                    params={"value": value},
                )
        return value


class Keccak256Filter(django_filters.Filter):
    field_class = Keccak256FieldForm



from logging import getLogger
from typing import Optional

from eth_account.signers.local import LocalAccount
from eth_typing import ChecksumAddress
from web3 import Web3
from web3.contract import Contract

from gnosis.eth import EthereumClient
from gnosis.eth.contracts import (
    get_paying_proxy_deployed_bytecode,
    get_proxy_1_0_0_deployed_bytecode,
    get_proxy_1_1_1_deployed_bytecode,
    get_proxy_1_1_1_mainnet_deployed_bytecode,
    get_proxy_1_3_0_deployed_bytecode,
    get_proxy_factory_contract,
    get_proxy_factory_V1_0_0_contract,
    get_proxy_factory_V1_1_1_contract,
)
from gnosis.eth.ethereum_client import EthereumTxSent
from gnosis.eth.utils import compare_byte_code

try:
    from functools import cache
except ImportError:
    from functools import lru_cache

    cache = lru_cache(maxsize=None)


logger = getLogger(__name__)


class ProxyFactory:
    def __init__(self, address: ChecksumAddress, ethereum_client: EthereumClient):
        assert Web3.isChecksumAddress(address), (
            "%s proxy factory address not valid" % address
        )

        self.address = address
        self.ethereum_client = ethereum_client
        self.w3 = ethereum_client.w3

    @staticmethod
    def _deploy_proxy_factory_contract(
        ethereum_client: EthereumClient,
        deployer_account: LocalAccount,
        contract: Contract,
    ) -> EthereumTxSent:
        tx = contract.constructor().buildTransaction({"from": deployer_account.address})

        tx_hash = ethereum_client.send_unsigned_transaction(
            tx, private_key=deployer_account.key
        )
        tx_receipt = ethereum_client.get_transaction_receipt(tx_hash, timeout=120)
        assert tx_receipt
        assert tx_receipt["status"]
        contract_address = tx_receipt["contractAddress"]
        logger.info(
            "Deployed and initialized Proxy Factory Contract=%s by %s",
            contract_address,
            deployer_account.address,
        )
        return EthereumTxSent(tx_hash, tx, contract_address)

    @classmethod
    def deploy_proxy_factory_contract(
        cls, ethereum_client: EthereumClient, deployer_account: LocalAccount
    ) -> EthereumTxSent:
        """
        Deploy proxy factory contract last version (v1.3.0)

        :param ethereum_client:
        :param deployer_account: Ethereum Account
        :return: deployed contract address
        """
        proxy_factory_contract = get_proxy_factory_contract(ethereum_client.w3)
        return cls._deploy_proxy_factory_contract(
            ethereum_client, deployer_account, proxy_factory_contract
        )

    @classmethod
    def deploy_proxy_factory_contract_v1_1_1(
        cls, ethereum_client: EthereumClient, deployer_account: LocalAccount
    ) -> EthereumTxSent:
        """
        Deploy proxy factory contract v1.1.1

        :param ethereum_client:
        :param deployer_account: Ethereum Account
        :return: deployed contract address
        """
        proxy_factory_contract = get_proxy_factory_V1_1_1_contract(ethereum_client.w3)
        return cls._deploy_proxy_factory_contract(
            ethereum_client, deployer_account, proxy_factory_contract
        )

    @classmethod
    def deploy_proxy_factory_contract_v1_0_0(
        cls, ethereum_client: EthereumClient, deployer_account: LocalAccount
    ) -> EthereumTxSent:
        """
        Deploy proxy factory contract v1.0.0

        :param ethereum_client:
        :param deployer_account: Ethereum Account
        :return: deployed contract address
        """
        proxy_factory_contract = get_proxy_factory_V1_0_0_contract(ethereum_client.w3)
        return cls._deploy_proxy_factory_contract(
            ethereum_client, deployer_account, proxy_factory_contract
        )

    def check_proxy_code(self, address: ChecksumAddress) -> bool:
        """
        Check if proxy is valid
        :param address: Ethereum address to check
        :return: True if proxy is valid, False otherwise
        """

        deployed_proxy_code = self.w3.eth.get_code(address)
        proxy_code_fns = (
            get_proxy_1_3_0_deployed_bytecode,
            get_proxy_1_1_1_deployed_bytecode,
            get_proxy_1_1_1_mainnet_deployed_bytecode,
            get_proxy_1_0_0_deployed_bytecode,
            get_paying_proxy_deployed_bytecode,
            self.get_proxy_runtime_code,
        )
        for proxy_code_fn in proxy_code_fns:
            if compare_byte_code(deployed_proxy_code, proxy_code_fn()):
                return True
        return False

    def deploy_proxy_contract(
        self,
        deployer_account: LocalAccount,
        master_copy: ChecksumAddress,
        initializer: bytes = b"",
        gas: Optional[int] = None,
        gas_price: Optional[int] = None,
    ) -> EthereumTxSent:
        """
        Deploy proxy contract via ProxyFactory using `createProxy` function
        :param deployer_account: Ethereum account
        :param master_copy: Address the proxy will point at
        :param initializer: Initializer
        :param gas: Gas
        :param gas_price: Gas Price
        :return: EthereumTxSent
        """
        proxy_factory_contract = self.get_contract()
        create_proxy_fn = proxy_factory_contract.functions.createProxy(
            master_copy, initializer
        )

        tx_parameters = {"from": deployer_account.address}
        contract_address = create_proxy_fn.call(tx_parameters)

        if gas_price is not None:
            tx_parameters["gasPrice"] = gas_price

        if gas is not None:
            tx_parameters["gas"] = gas

        tx = create_proxy_fn.buildTransaction(tx_parameters)
        # Auto estimation of gas does not work. We use a little more gas just in case
        tx["gas"] = tx["gas"] + 50000
        tx_hash = self.ethereum_client.send_unsigned_transaction(
            tx, private_key=deployer_account.key
        )
        return EthereumTxSent(tx_hash, tx, contract_address)

    def deploy_proxy_contract_with_nonce(
        self,
        deployer_account: LocalAccount,
        master_copy: ChecksumAddress,
        initializer: bytes,
        salt_nonce: int,
        user_account: str,
        gas: Optional[int] = None,
        gas_price: Optional[int] = None,
        nonce: Optional[int] = None,
    ) -> EthereumTxSent:
        """
        Deploy proxy contract via Proxy Factory using `createProxyWithNonce` (create2)

        :param deployer_account: Ethereum account
        :param master_copy: Address the proxy will point at
        :param initializer: Data for safe creation
        :param salt_nonce: Uint256 for `create2` salt
        :param gas: Gas
        :param gas_price: Gas Price
        :param nonce: Nonce
        :return: Tuple(tx-hash, tx, deployed contract address)
        """
        proxy_factory_contract = self.get_contract()
        create_proxy_fn = proxy_factory_contract.functions.createProxyWithNonce(
            master_copy, initializer, salt_nonce, user_account
        )

        tx_parameters = {"from": deployer_account.address}
        contract_address = create_proxy_fn.call(tx_parameters)

        if gas_price is not None:
            tx_parameters["gasPrice"] = gas_price

        if gas is not None:
            tx_parameters["gas"] = gas

        if nonce is not None:
            tx_parameters["nonce"] = nonce

        tx = create_proxy_fn.buildTransaction(tx_parameters)
        # Auto estimation of gas does not work. We use a little more gas just in case
        tx["gas"] = tx["gas"] + 50000
        tx_hash = self.ethereum_client.send_unsigned_transaction(
            tx, private_key=deployer_account.key
        )
        return EthereumTxSent(tx_hash, tx, contract_address)

    def get_contract(self, address: Optional[ChecksumAddress] = None):
        address = address or self.address
        return get_proxy_factory_contract(self.ethereum_client.w3, address)

    @cache
    def get_proxy_runtime_code(self, address: Optional[ChecksumAddress] = None):
        """
        Get runtime code for current proxy factory
        """
        address = address or self.address
        return self.get_contract(address=address).functions.proxyRuntimeCode().call()



cream_ctoken_abi = [
    {
        "inputs": [
            {"internalType": "address", "name": "underlying_", "type": "address"},
            {
                "internalType": "contract ComptrollerInterface",
                "name": "comptroller_",
                "type": "address",
            },
            {
                "internalType": "contract InterestRateModel",
                "name": "interestRateModel_",
                "type": "address",
            },
            {
                "internalType": "uint256",
                "name": "initialExchangeRateMantissa_",
                "type": "uint256",
            },
            {"internalType": "string", "name": "name_", "type": "string"},
            {"internalType": "string", "name": "symbol_", "type": "string"},
            {"internalType": "uint8", "name": "decimals_", "type": "uint8"},
            {"internalType": "address payable", "name": "admin_", "type": "address"},
            {"internalType": "address", "name": "implementation_", "type": "address"},
            {
                "internalType": "bytes",
                "name": "becomeImplementationData",
                "type": "bytes",
            },
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "constructor",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "cashPrior",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "interestAccumulated",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "borrowIndex",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "totalBorrows",
                "type": "uint256",
            },
        ],
        "name": "AccrueInterest",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "owner",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "spender",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "amount",
                "type": "uint256",
            },
        ],
        "name": "Approval",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "address",
                "name": "borrower",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "borrowAmount",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "accountBorrows",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "totalBorrows",
                "type": "uint256",
            },
        ],
        "name": "Borrow",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "error",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "info",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "detail",
                "type": "uint256",
            },
        ],
        "name": "Failure",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "address",
                "name": "liquidator",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "address",
                "name": "borrower",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "repayAmount",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "address",
                "name": "cTokenCollateral",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "seizeTokens",
                "type": "uint256",
            },
        ],
        "name": "LiquidateBorrow",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "address",
                "name": "minter",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "mintAmount",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "mintTokens",
                "type": "uint256",
            },
        ],
        "name": "Mint",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "address",
                "name": "oldAdmin",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "address",
                "name": "newAdmin",
                "type": "address",
            },
        ],
        "name": "NewAdmin",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "contract ComptrollerInterface",
                "name": "oldComptroller",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "contract ComptrollerInterface",
                "name": "newComptroller",
                "type": "address",
            },
        ],
        "name": "NewComptroller",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "address",
                "name": "oldImplementation",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "address",
                "name": "newImplementation",
                "type": "address",
            },
        ],
        "name": "NewImplementation",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "contract InterestRateModel",
                "name": "oldInterestRateModel",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "contract InterestRateModel",
                "name": "newInterestRateModel",
                "type": "address",
            },
        ],
        "name": "NewMarketInterestRateModel",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "address",
                "name": "oldPendingAdmin",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "address",
                "name": "newPendingAdmin",
                "type": "address",
            },
        ],
        "name": "NewPendingAdmin",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "oldReserveFactorMantissa",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "newReserveFactorMantissa",
                "type": "uint256",
            },
        ],
        "name": "NewReserveFactor",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "address",
                "name": "redeemer",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "redeemAmount",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "redeemTokens",
                "type": "uint256",
            },
        ],
        "name": "Redeem",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "address",
                "name": "payer",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "address",
                "name": "borrower",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "repayAmount",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "accountBorrows",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "totalBorrows",
                "type": "uint256",
            },
        ],
        "name": "RepayBorrow",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "address",
                "name": "benefactor",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "addAmount",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "newTotalReserves",
                "type": "uint256",
            },
        ],
        "name": "ReservesAdded",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "address",
                "name": "admin",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "reduceAmount",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "newTotalReserves",
                "type": "uint256",
            },
        ],
        "name": "ReservesReduced",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "from",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "to",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "amount",
                "type": "uint256",
            },
        ],
        "name": "Transfer",
        "type": "event",
    },
    {"payable": True, "stateMutability": "payable", "type": "fallback"},
    {
        "constant": False,
        "inputs": [],
        "name": "_acceptAdmin",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [{"internalType": "uint256", "name": "addAmount", "type": "uint256"}],
        "name": "_addReserves",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "uint256", "name": "reduceAmount", "type": "uint256"}
        ],
        "name": "_reduceReserves",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "contract ComptrollerInterface",
                "name": "newComptroller",
                "type": "address",
            }
        ],
        "name": "_setComptroller",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "implementation_", "type": "address"},
            {"internalType": "bool", "name": "allowResign", "type": "bool"},
            {
                "internalType": "bytes",
                "name": "becomeImplementationData",
                "type": "bytes",
            },
        ],
        "name": "_setImplementation",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "contract InterestRateModel",
                "name": "newInterestRateModel",
                "type": "address",
            }
        ],
        "name": "_setInterestRateModel",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "address payable",
                "name": "newPendingAdmin",
                "type": "address",
            }
        ],
        "name": "_setPendingAdmin",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "uint256",
                "name": "newReserveFactorMantissa",
                "type": "uint256",
            }
        ],
        "name": "_setReserveFactor",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "accrualBlockNumber",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [],
        "name": "accrueInterest",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "admin",
        "outputs": [{"internalType": "address payable", "name": "", "type": "address"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [
            {"internalType": "address", "name": "owner", "type": "address"},
            {"internalType": "address", "name": "spender", "type": "address"},
        ],
        "name": "allowance",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "spender", "type": "address"},
            {"internalType": "uint256", "name": "amount", "type": "uint256"},
        ],
        "name": "approve",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [{"internalType": "address", "name": "owner", "type": "address"}],
        "name": "balanceOf",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [{"internalType": "address", "name": "owner", "type": "address"}],
        "name": "balanceOfUnderlying",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "uint256", "name": "borrowAmount", "type": "uint256"}
        ],
        "name": "borrow",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [{"internalType": "address", "name": "account", "type": "address"}],
        "name": "borrowBalanceCurrent",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [{"internalType": "address", "name": "account", "type": "address"}],
        "name": "borrowBalanceStored",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "borrowIndex",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "borrowRatePerBlock",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "comptroller",
        "outputs": [
            {
                "internalType": "contract ComptrollerInterface",
                "name": "",
                "type": "address",
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "decimals",
        "outputs": [{"internalType": "uint8", "name": "", "type": "uint8"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [{"internalType": "bytes", "name": "data", "type": "bytes"}],
        "name": "delegateToImplementation",
        "outputs": [{"internalType": "bytes", "name": "", "type": "bytes"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [{"internalType": "bytes", "name": "data", "type": "bytes"}],
        "name": "delegateToViewImplementation",
        "outputs": [{"internalType": "bytes", "name": "", "type": "bytes"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [],
        "name": "exchangeRateCurrent",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "exchangeRateStored",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [{"internalType": "address", "name": "account", "type": "address"}],
        "name": "getAccountSnapshot",
        "outputs": [
            {"internalType": "uint256", "name": "", "type": "uint256"},
            {"internalType": "uint256", "name": "", "type": "uint256"},
            {"internalType": "uint256", "name": "", "type": "uint256"},
            {"internalType": "uint256", "name": "", "type": "uint256"},
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "getCash",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "implementation",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "interestRateModel",
        "outputs": [
            {
                "internalType": "contract InterestRateModel",
                "name": "",
                "type": "address",
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "isCToken",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "borrower", "type": "address"},
            {"internalType": "uint256", "name": "repayAmount", "type": "uint256"},
            {
                "internalType": "contract CTokenInterface",
                "name": "cTokenCollateral",
                "type": "address",
            },
        ],
        "name": "liquidateBorrow",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "uint256", "name": "mintAmount", "type": "uint256"}
        ],
        "name": "mint",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "name",
        "outputs": [{"internalType": "string", "name": "", "type": "string"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "pendingAdmin",
        "outputs": [{"internalType": "address payable", "name": "", "type": "address"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "uint256", "name": "redeemTokens", "type": "uint256"}
        ],
        "name": "redeem",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "uint256", "name": "redeemAmount", "type": "uint256"}
        ],
        "name": "redeemUnderlying",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "uint256", "name": "repayAmount", "type": "uint256"}
        ],
        "name": "repayBorrow",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "reserveFactorMantissa",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "liquidator", "type": "address"},
            {"internalType": "address", "name": "borrower", "type": "address"},
            {"internalType": "uint256", "name": "seizeTokens", "type": "uint256"},
        ],
        "name": "seize",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "supplyRatePerBlock",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "symbol",
        "outputs": [{"internalType": "string", "name": "", "type": "string"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "totalBorrows",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [],
        "name": "totalBorrowsCurrent",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "totalReserves",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "totalSupply",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "dst", "type": "address"},
            {"internalType": "uint256", "name": "amount", "type": "uint256"},
        ],
        "name": "transfer",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "src", "type": "address"},
            {"internalType": "address", "name": "dst", "type": "address"},
            {"internalType": "uint256", "name": "amount", "type": "uint256"},
        ],
        "name": "transferFrom",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "underlying",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
]



# flake8: noqa F401
from .ethereum_client import (
    EthereumClient,
    EthereumClientProvider,
    FromAddressNotFound,
    GasLimitExceeded,
    InsufficientFunds,
    InvalidNonce,
    NonceTooHigh,
    NonceTooLow,
    ReplacementTransactionUnderpriced,
    SenderAccountNotFoundInNode,
    TransactionAlreadyImported,
    TransactionQueueLimitReached,
    TxSpeed,
    UnknownAccount,
)
from .ethereum_network import EthereumNetwork, EthereumNetworkNotSupported
from .exceptions import InvalidERC20Info, InvalidERC721Info, ParityTraceDecodeException



from django.test import TestCase

from eth_account import Account
from web3 import Web3

from gnosis.eth.constants import NULL_ADDRESS

from ..signatures import get_signing_address


class TestSafeSignature(TestCase):
    def test_get_signing_address(self):
        account = Account.create()
        # Random hash
        random_hash = Web3.keccak(text="tanxugueiras")
        signature = account.signHash(random_hash)
        self.assertEqual(
            get_signing_address(random_hash, signature.v, signature.r, signature.s),
            account.address,
        )

        # Invalid signature will return the `NULL_ADDRESS`
        self.assertEqual(
            get_signing_address(random_hash, signature.v - 8, signature.r, signature.s),
            NULL_ADDRESS,
        )
